"""State module for managing subnetworks."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools.typing import Computed

from idem_gcp.tool.gcp.utils import regional_absent

# prevent commit hook from removing the import
absent = regional_absent

from idem_gcp.tool.gcp.state_operation_utils import StateOperations

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    region: str = None,
    project: str = None,
    request_id: str = None,
    enable_flow_logs: bool = None,
    private_ip_google_access: bool = None,
    description: str = None,
    network: str = None,
    stack_type: str = None,
    log_config: make_dataclass(
        "SubnetworkLogConfig",
        [
            ("aggregation_interval", str, field(default=None)),
            ("filter_expr", str, field(default=None)),
            ("enable", bool, field(default=None)),
            ("flow_sampling", float, field(default=None)),
            ("metadata_fields", List[str], field(default=None)),
            ("metadata", str, field(default=None)),
        ],
    ) = None,
    role: str = None,
    ipv6_access_type: str = None,
    fingerprint: str = None,
    secondary_ip_ranges: List[
        make_dataclass(
            "SubnetworkSecondaryRange",
            [
                ("ip_cidr_range", str, field(default=None)),
                ("range_name", str, field(default=None)),
            ],
        )
    ] = None,
    purpose: str = None,
    private_ipv6_google_access: str = None,
    ip_cidr_range: str = None,
    drain_timeout_seconds: int = None,
    resource_id: str = None,
    id_: Computed[str] = None,
) -> Dict[str, Any]:
    r"""Creates a subnetwork in the specified project using the data included in the request.

    Args:
        name(str):
            An Idem name of the resource.

        project(str, Optional):
            Project ID for this request.

        region(str):
            Name of the region scoping this request.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        enable_flow_logs(bool, Optional):
            Whether to enable flow logging for this subnetwork. If this field is not explicitly set, it will not appear in get listings. If not set the default behavior is determined by the org policy, if there is no org policy specified, then it will default to disabled. This field isn't supported with the purpose field set to INTERNAL_HTTPS_LOAD_BALANCER. Defaults to None.

        private_ip_google_access(bool, Optional):
            Whether the VMs in this subnet can access Google services without assigned external IP addresses. This field can be both set at resource creation time and updated using setPrivateIpGoogleAccess. Defaults to None.

        description(str, Optional):
            An optional description of this resource. Provide this property when you create the resource. This field can be set only at resource creation time. Defaults to None.

        network(str, Optional):
            The URL of the network to which this subnetwork belongs, provided by the client when initially creating the subnetwork. This field can be set only at resource creation time. Defaults to None.

        stack_type(str, Optional):
            The stack type for the subnet. If set to IPV4_ONLY, new VMs in the subnet are assigned IPv4 addresses only. If set to IPV4_IPV6, new VMs in the subnet can be assigned both IPv4 and IPv6 addresses. If not specified, IPV4_ONLY is used. This field can be both set at resource creation time and updated using patch.
            Enum type. Allowed values:
            "IPV4_IPV6" - New VMs in this subnet can have both IPv4 and IPv6 addresses.
            "IPV4_ONLY" - New VMs in this subnet will only be assigned IPv4 addresses. Defaults to None.

        log_config(Dict[str, Any], Optional):
            This field denotes the VPC flow logging options for this subnetwork. If logging is enabled, logs are exported to Cloud Logging.
            SubnetworkLogConfig: The available logging options for this subnetwork. Defaults to None.
            * aggregation_interval (str, Optional): Can only be specified if VPC flow logging for this subnetwork is enabled. Toggles the aggregation interval for collecting flow logs. Increasing the interval time will reduce the amount of generated flow logs for long lasting connections. Default is an interval of 5 seconds per connection.
                Enum type. Allowed values:
                "INTERVAL_10_MIN"
                "INTERVAL_15_MIN"
                "INTERVAL_1_MIN"
                "INTERVAL_30_SEC"
                "INTERVAL_5_MIN"
                "INTERVAL_5_SEC"
            * filter_expr (str, Optional): Can only be specified if VPC flow logs for this subnetwork is enabled. The filter expression is used to define which VPC flow logs should be exported to Cloud Logging.
            * enable (bool, Optional): Whether to enable flow logging for this subnetwork. If this field is not explicitly set, it will not appear in get listings. If not set the default behavior is determined by the org policy, if there is no org policy specified, then it will default to disabled.
            * flow_sampling (float, Optional): Can only be specified if VPC flow logging for this subnetwork is enabled. The value of the field must be in [0, 1]. Set the sampling rate of VPC flow logs within the subnetwork where 1.0 means all collected logs are reported and 0.0 means no logs are reported. Default is 0.5 unless otherwise specified by the org policy, which means half of all collected logs are reported.
            * metadata_fields (List[str], Optional): Can only be specified if VPC flow logs for this subnetwork is enabled and "metadata" was set to CUSTOM_METADATA.
            * metadata (str, Optional): Can only be specified if VPC flow logs for this subnetwork is enabled. Configures whether all, none or a subset of metadata fields should be added to the reported VPC flow logs. Default is EXCLUDE_ALL_METADATA.
                Enum type. Allowed values:
                "CUSTOM_METADATA"
                "EXCLUDE_ALL_METADATA"
                "INCLUDE_ALL_METADATA"

        role(str, Optional):
            The role of subnetwork. Currently, this field is only used when purpose = INTERNAL_HTTPS_LOAD_BALANCER. The value can be set to ACTIVE or BACKUP. An ACTIVE subnetwork is one that is currently being used for Internal HTTP(S) Load Balancing. A BACKUP subnetwork is one that is ready to be promoted to ACTIVE or is currently draining. This field can be updated with a patch request.
            Enum type. Allowed values:
            "ACTIVE" - The ACTIVE subnet that is currently used.
            "BACKUP" - The BACKUP subnet that could be promoted to ACTIVE. Defaults to None.

        ipv6_access_type(str, Optional):
            The access type of IPv6 address this subnet holds. It's immutable and can only be specified during creation or the first time the subnet is updated into IPV4_IPV6 dual stack.
            Enum type. Allowed values:
            "EXTERNAL" - VMs on this subnet will be assigned IPv6 addresses that are accessible via the Internet, as well as the VPC network.
            "INTERNAL" - VMs on this subnet will be assigned IPv6 addresses that are only accessible over the VPC network. Defaults to None.

        fingerprint(str, Optional):
            Fingerprint of this resource. A hash of the contents stored in this object. This field is used in optimistic locking. This field will be ignored when inserting a Subnetwork. An up-to-date fingerprint must be provided in order to update the Subnetwork, otherwise the request will fail with error 412 conditionNotMet. To see the latest fingerprint, make a get() request to retrieve a Subnetwork. Defaults to None.

        secondary_ip_ranges(List[Dict[str, Any]], Optional):
            An array of configurations for secondary IP ranges for VM instances contained in this subnetwork. The primary IP of such VM must belong to the primary ipCidrRange of the subnetwork. The alias IPs may belong to either primary or secondary ranges. This field can be updated with a patch request. Defaults to None.
            * ip_cidr_range (str, optional): The range of IP addresses belonging to this subnetwork secondary range. Provide this property when you create the subnetwork. Ranges must be unique and non-overlapping with all primary and secondary IP ranges within a network. Only IPv4 is supported. The range can be any range listed in the Valid ranges list.
            * range_name (str, optional): The name associated with this subnetwork secondary range, used when adding an alias IP range to a VM instance. The name must be 1-63 characters long, and comply with RFC1035. The name must be unique within the subnetwork.

        purpose(str, Optional):
            The purpose of the resource. This field can be either PRIVATE_RFC_1918 or INTERNAL_HTTPS_LOAD_BALANCER. A subnetwork with purpose set to INTERNAL_HTTPS_LOAD_BALANCER is a user-created subnetwork that is reserved for Internal HTTP(S) Load Balancing. If unspecified, the purpose defaults to PRIVATE_RFC_1918. The enableFlowLogs field isn't supported with the purpose field set to INTERNAL_HTTPS_LOAD_BALANCER.
            Enum type. Allowed values:
            "INTERNAL_HTTPS_LOAD_BALANCER" - Subnet reserved for Internal HTTP(S) Load Balancing.
            "PRIVATE" - Regular user created or automatically created subnet.
            "PRIVATE_RFC_1918" - Regular user created or automatically created subnet.
            "PRIVATE_SERVICE_CONNECT" - Subnetworks created for Private Service Connect in the producer network.
            "REGIONAL_MANAGED_PROXY" - Subnetwork used for Regional Internal/External HTTP(S) Load Balancing. Defaults to None.

        private_ipv6_google_access(str, Optional):
            This field is for internal use. This field can be both set at resource creation time and updated using patch.
            Enum type. Allowed values:
            "DISABLE_GOOGLE_ACCESS" - Disable private IPv6 access to/from Google services.
            "ENABLE_BIDIRECTIONAL_ACCESS_TO_GOOGLE" - Bidirectional private IPv6 access to/from Google services.
            "ENABLE_OUTBOUND_VM_ACCESS_TO_GOOGLE" - Outbound private IPv6 access from VMs in this subnet to Google services. Defaults to None.

        ip_cidr_range(str, Optional):
            The range of internal addresses that are owned by this subnetwork. Provide this property when you create the subnetwork. For example, 10.0.0.0/8 or 100.64.0.0/10. Ranges must be unique and non-overlapping within a network. Only IPv4 is supported. This field is set at resource creation time. The range can be any range listed in the Valid ranges list. The range can be expanded after creation using expandIpCidrRange. Defaults to None.

        drain_timeout_seconds(int, Optional):
            The drain timeout specifies the upper bound in seconds on the amount of time allowed to drain connections from the current ACTIVE subnetwork to the current BACKUP subnetwork. The drain timeout is only applicable when the following conditions are true: - the subnetwork being patched has purpose = INTERNAL_HTTPS_LOAD_BALANCER - the subnetwork being patched has role = BACKUP - the patch request is setting the role to ACTIVE. Note that after this patch operation the roles of the ACTIVE and BACKUP subnetworks will be swapped. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        id_(str, optional): The unique identifier for the resource. This identifier is defined by the server. Read-only property

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_resource_name:
              gcp.compute.subnetwork.present:
                - name: value
                - region: value
                - project: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if ctx.get("wrapper_result"):
        result = ctx.get("wrapper_result")

    request_body = {
        "log_config": log_config,
        "purpose": purpose,
        "private_ipv6_google_access": private_ipv6_google_access,
        "private_ip_google_access": private_ip_google_access,
        "role": role,
        "ipv6_access_type": ipv6_access_type,
        "name": name,
        "ip_cidr_range": ip_cidr_range,
        "description": description,
        "enable_flow_logs": enable_flow_logs,
        "stack_type": stack_type,
        "secondary_ip_ranges": secondary_ip_ranges,
        "network": network,
        "id_": id_,
    }

    operation = None
    plan_state = {"resource_id": resource_id, **request_body}
    plan_state = {k: v for (k, v) in plan_state.items() if v is not None}

    if result["old_state"]:
        resource_id = result["old_state"].get("resource_id", None)
        # Update subnetwork
        patch_operations_dict = {
            "ip_cidr_range": (
                hub.tool.gcp.compute.subnetwork.update_ip_cidr_range,
                (ctx, result["old_state"], ip_cidr_range),
                True,
            ),
        }

        state_operations = StateOperations(
            hub, "compute.subnetwork", patch_operations_dict, result, request_body
        )

        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            {
                "resource_id": resource_id,
                **request_body,
            },
            "compute.subnetwork",
            additional_exclude_paths=list(patch_operations_dict.keys()),
        )

        if not changes and not any(state_operations.changed_properties_dict.values()):
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(
                    "gcp.compute.subnetwork", name
                )
            )
            result["new_state"] = result["old_state"]
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    "gcp.compute.subnetwork", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                plan_state
            )
            return result

        state_operations_ret = await state_operations.run_operations()

        if not state_operations_ret["result"]:
            result["comment"] += state_operations_ret["comment"]
            result["result"] = False
            return result

        if changes:
            if any(state_operations.changed_properties_dict.values()):
                get_ret = await hub.exec.gcp.compute.subnetwork.get(
                    ctx, name=name, resource_id=resource_id
                )

                if not get_ret["result"] and not get_ret["ret"]:
                    result["result"] = False
                    result["comment"] += get_ret["comment"]
                    return result

                for k in patch_operations_dict.keys():
                    request_body[k] = get_ret["ret"].get(k)

                request_body["fingerprint"] = get_ret["ret"].get("fingerprint")

            body = (
                {**request_body, "fingerprint": result["old_state"].get("fingerprint")}
                if not request_body.get("fingerprint")
                else {**request_body}
            )
            body.pop("ip_cidr_range")
            body.pop("network")
            state_operations.pre_process_resource_body(body)
            update = await hub.exec.gcp_api.client.compute.subnetwork.patch(
                ctx, resource_id=resource_id, body=body
            )
            if not update["result"]:
                result["result"] = False
                result["comment"] += update["comment"]
                return result

            if hub.tool.gcp.operation_utils.is_operation(update["ret"]):
                operation = update["ret"]

        if not changes and any(state_operations.changed_properties_dict.values()):
            get_ret = await hub.exec.gcp.compute.subnetwork.get(
                ctx, name=name, resource_id=resource_id
            )
            if not get_ret["result"] and not get_ret["ret"]:
                result["result"] = False
                result["comment"] += get_ret["comment"]
                return result

            result["new_state"] = get_ret["ret"]
            result["comment"].append(
                hub.tool.gcp.comment_utils.update_comment(
                    "gcp.compute.subnetwork", name
                )
            )
            return result
    else:
        # Create subnetwork
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.subnetwork", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                plan_state
            )
            result["new_state"][
                "resource_id"
            ] = hub.tool.gcp.resource_prop_utils.construct_resource_id(
                "compute.subnetwork", {**locals(), "subnetwork": name}
            )
            return result

        create = await hub.exec.gcp_api.client.compute.subnetwork.insert(
            ctx, project=project, region=region, body=request_body
        )
        if not create["result"]:
            result["result"] = False
            result["comment"] += create["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(create["ret"]):
            operation = create["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.region_operation"
        )
        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of subnetworks available to the specified project.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.subnetwork
    """
    result = {}

    list_result = await hub.exec.gcp.compute.subnetwork.list(
        ctx, project=ctx.acct.project_id
    )

    if not list_result["result"]:
        hub.log.debug(f"Could not describe subnetworks {list_result['comment']}")
        return {}

    for resource in list_result["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.subnetwork.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""Idem plugin for the Google API Python client library.

Google API Python client library
:copyright: Copyright (c) 2020-2022 VMware, Inc. All Rights Reserved.
:license: SPDX-License-Identifier: Apache-2.0

This file declares and implements a class to wrap and make easy access to
Google's discovery based APIs. Additionally, it declares and implements the
entirety of the Plugin Oriented Programming (POP) exec Sub for Google API
Management via POP ReverseSub.

The basic operating model is building Subs as needed to match the nature of a
call to an exec. For example, a call like:

    gcp.compute.instance.get(ctx, ...)

Those are ultimately implemented by calls to the _GCPApi class runnable.

will work, even though no exec subdirectories exist to match that call path.
Instead, the API is dynamically discovered and built to match and tie
them to the appropriate Google Python SDK wrappers within the tool Sub of
this project.
"""
import time
from inspect import signature
from socket import timeout
from urllib.parse import urlparse

try:
    from googleapiclient import discovery
    from googleapiclient.discovery_cache import base as base_cache

    import idem_gcp.helpers.exc as idemexc

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def __init__(hub):
    # hub.pop.sub.load_subdirs(hub.tool.gcp, recurse=True)
    hub.pop.sub.add(dyne_name="metadata")
    hub.pop.sub.load_subdirs(hub.metadata, recurse=True)
    """
    Initialize default API versions and setup the root _GCPApi object.
    :param hub: The Idem hub to use for (attach to) the root of the
    GCPApi API tree that gets dynamically built.
    """
    # hub.pop.config.load(["idem", "acct", "idem_gcp", "rend", "evbus"], "idem", parse_cli=True)
    hub.tool.gcp.DEFAULT_VERSIONS = {
        "abusiveexperiencereport": "v1",
        "acceleratedmobilepageurl": "v1",
        "accessapproval": "v1",
        "accesscontextmanager": "v1",
        "adexchangebuyer2": "v2beta1",
        "adexperiencereport": "v1",
        #'admin': ['datatransfer_v1', 'directory_v1', 'reports_v1'],
        "admob": "v1",
        "adsense": "v2",
        "adsensehost": "v4.1",
        "alertcenter": "v1beta1",
        "analytics": "v3",
        "analyticsadmin": "v1alpha",
        "analyticsdata": "v1beta",
        "analyticshub": "v1beta1",
        "analyticsreporting": "v4",
        "androiddeviceprovisioning": "v1",
        "androidenterprise": "v1",
        "androidmanagement": "v1",
        "androidpublisher": "v3",
        "apigateway": "v1",
        "apigee": "v1",
        "apigeeregistry": "v1",
        "apikeys": "v2",
        "appengine": "v1",
        "area120tables": "v1alpha1",
        "artifactregistry": "v1",
        "assuredworkloads": "v1",
        "authorizedbuyersmarketplace": "v1",
        "baremetalsolution": "v2",
        "beyondcorp": "v1alpha",
        "bigquery": "v2",
        "bigqueryconnection": "v1beta1",
        "bigquerydatatransfer": "v1",
        "bigqueryreservation": "v1",
        "bigtableadmin": "v2",
        "billingbudgets": "v1",
        "binaryauthorization": "v1",
        "blogger": "v3",
        "books": "v1",
        "calendar": "v3",
        "certificatemanager": "v1",
        "chat": "v1",
        "chromemanagement": "v1",
        "chromepolicy": "v1",
        "chromeuxreport": "v1",
        "civicinfo": "v2",
        "classroom": "v1",
        "cloudasset": "v1",
        "cloudbilling": "v1",
        "cloudbuild": "v1",
        "cloudchannel": "v1",
        "clouddebugger": "v2",
        "clouddeploy": "v1",
        "clouderrorreporting": "v1beta1",
        "cloudfunctions": "v2",
        "cloudidentity": "v1",
        "cloudiot": "v1",
        "cloudkms": "v1",
        "cloudprofiler": "v2",
        "cloudresourcemanager": "v3",
        "cloudscheduler": "v1",
        "cloudsearch": "v1",
        "cloudshell": "v1",
        "cloudsupport": "v2beta",
        "cloudtasks": "v2",
        "cloudtrace": "v2",
        "composer": "v1",
        "compute": "v1",
        "connectors": "v1",
        "contactcenterinsights": "v1",
        "container": "v1",
        "containeranalysis": "v1",
        "content": "v2.1",
        "customsearch": "v1",
        "datacatalog": "v1",
        "dataflow": "v1b3",
        "datafusion": "v1",
        "datalabeling": "v1beta1",
        "datamigration": "v1",
        "datapipelines": "v1",
        "dataplex": "v1",
        "dataproc": "v1",
        "datastore": "v1",
        "datastream": "v1",
        "deploymentmanager": "v2",
        "dfareporting": "v3.5",
        "dialogflow": "v3",
        "digitalassetlinks": "v1",
        "discovery": "v1",
        "displayvideo": "v1",
        "dlp": "v2",
        "dns": "v1",
        "docs": "v1",
        "documentai": "v1",
        "domains": "v1",
        "domainsrdap": "v1",
        "doubleclickbidmanager": "v1.1",
        "doubleclicksearch": "v2",
        "drive": "v3",
        "driveactivity": "v2",
        "essentialcontacts": "v1",
        "eventarc": "v1",
        "factchecktools": "v1alpha1",
        "fcm": "v1",
        "fcmdata": "v1beta1",
        "file": "v1",
        "firebase": "v1beta1",
        "firebaseappcheck": "v1",
        "firebasedatabase": "v1beta",
        "firebasedynamiclinks": "v1",
        "firebasehosting": "v1",
        "firebaseml": "v1",
        "firebaserules": "v1",
        "firebasestorage": "v1beta",
        "firestore": "v1",
        "fitness": "v1",
        "forms": "v1",
        "games": "v1",
        "gamesConfiguration": "v1configuration",
        "gamesManagement": "v1management",
        "gameservices": "v1",
        "genomics": "v2alpha1",
        "gkebackup": "v1",
        "gkehub": "v1",
        "gmail": "v1",
        "gmailpostmastertools": "v1",
        "groupsmigration": "v1",
        "groupssettings": "v1",
        "healthcare": "v1",
        "homegraph": "v1",
        "iam": "v1",
        "iamcredentials": "v1",
        "iap": "v1",
        "ideahub": "v1beta",
        "identitytoolkit": "v3",
        "ids": "v1",
        "indexing": "v3",
        "jobs": "v4",
        "keep": "v1",
        "kgsearch": "v1",
        "language": "v1",
        "libraryagent": "v1",
        "licensing": "v1",
        "lifesciences": "v2beta",
        "localservices": "v1",
        "logging": "v2",
        "managedidentities": "v1",
        "manufacturers": "v1",
        "memcache": "v1",
        "metastore": "v1beta",
        "ml": "v1",
        "monitoring": "v3",
        "mybusinessaccountmanagement": "v1",
        "mybusinessbusinesscalls": "v1",
        "mybusinessbusinessinformation": "v1",
        "mybusinesslodging": "v1",
        "mybusinessnotifications": "v1",
        "mybusinessplaceactions": "v1",
        "mybusinessqanda": "v1",
        "mybusinessverifications": "v1",
        "networkconnectivity": "v1",
        "networkmanagement": "v1",
        "networksecurity": "v1",
        "networkservices": "v1",
        "notebooks": "v1",
        "oauth2": "v2",
        "ondemandscanning": "v1",
        "orgpolicy": "v2",
        "osconfig": "v1",
        "oslogin": "v1",
        "pagespeedonline": "v5",
        "paymentsresellersubscription": "v1",
        "people": "v1",
        "playcustomapp": "v1",
        "playdeveloperreporting": "v1beta1",
        "playintegrity": "v1",
        "policyanalyzer": "v1",
        "policysimulator": "v1",
        "policytroubleshooter": "v1",
        "poly": "v1",
        "privateca": "v1",
        "prod_tt_sasportal": "v1alpha1",
        "pubsub": "v1",
        "pubsublite": "v1",
        "realtimebidding": "v1",
        "recaptchaenterprise": "v1",
        "recommendationengine": "v1beta1",
        "recommender": "v1",
        "redis": "v1",
        "reseller": "v1",
        "resourcesettings": "v1",
        "retail": "v2",
        "run": "v2",
        "runtimeconfig": "v1",
        "safebrowsing": "v4",
        "sasportal": "v1alpha1",
        "script": "v1",
        "searchconsole": "v1",
        "secretmanager": "v1",
        "securitycenter": "v1",
        "serviceconsumermanagement": "v1",
        "servicecontrol": "v2",
        "servicedirectory": "v1",
        "servicemanagement": "v1",
        "servicenetworking": "v1",
        "serviceusage": "v1",
        "sheets": "v4",
        "siteVerification": "v1",
        "slides": "v1",
        "smartdevicemanagement": "v1",
        "sourcerepo": "v1",
        "spanner": "v1",
        "speech": "v1",
        "sqladmin": "v1",
        "storage": "v1",
        "storagetransfer": "v1",
        "streetviewpublish": "v1",
        "sts": "v1",
        "tagmanager": "v2",
        "tasks": "v1",
        "testing": "v1",
        "texttospeech": "v1",
        "toolresults": "v1beta3",
        "tpu": "v1",
        "trafficdirector": "v2",
        "transcoder": "v1",
        "translate": "v3",
        "vault": "v1",
        "verifiedaccess": "v2",
        "versionhistory": "v1",
        "videointelligence": "v1",
        "vision": "v1",
        "vmmigration": "v1",
        "webfonts": "v1",
        "webrisk": "v1",
        "websecurityscanner": "v1",
        "workflowexecutions": "v1",
        "workflows": "v1",
        "youtube": "v3",
        "youtubeAnalytics": "v2",
        "youtubereporting": "v1",
    }
    hub.tool.gcp.API = _GCPApi(hub, "gcp")


class _MemoryCache(base_cache.Cache):
    """Class to cover for the lack of discovery client file_cache when using recent oauth2client releases.

    See `Github commit
    <https://github.com/GoogleCloudPlatform/python-docs-samples/commit/22788da481a8441500203ccc7fbf37cd9fcafa3b>`_.
    """

    _CACHE = {}

    def get(self, url):
        return _MemoryCache._CACHE.get(url)

    def set(self, url, content):
        _MemoryCache._CACHE[url] = content


class _GCPApi:
    """Class to represent (wrap) `Google's discovery based APIs <https://github.com/googleapis/google-api-python-client>`_.

    Each instance represents a single service or (sub)resource in a way that
    enables writing code such as::

        my_gcpapi.iam.roles.projects.list(ctx, *args, **kwargs)

    in order to call specific apis. The class is callable and creates the
    Google API objects as needed to resolve such code.
    """

    _DISCOVERY_CACHE = _MemoryCache()

    # Operational Flags
    WAIT_ON_OPERATIONS = False

    # Call return codes
    E_NONE = 0
    E_NOTFOUND = -1
    E_JSONERROR = -2
    E_MEDIAERROR = -3
    E_MUTUALTLSCHANNELERROR = -4
    E_UNACCEPTABLEMIMETYPEERROR = -5

    def __init__(
        self, hub, resource_type: str, parent=None, cache: base_cache.Cache = None
    ):
        """Initializes the instance.

        :param hub: The Idem hub to use for related calls (e.g., logging).
        :param resource_type: The name of the service or resource to wrap.
        :type resource_type: str
        :param parent: The _GCPApi parent for this service or resource.
            If parent is None, then it the instance is considered a 'cloud'
            level API interface. For example, a resource_type of 'gcp' can be
            set with a parent of None to signify this is the root of the
            cloud API tree. Child instances would be initialized, for example,
            with resource_type as 'iam' and parent set to the _GCPApi root
            instance or some intermediate GCP API Resource (wrapper) instance.

            (default is None)
        :type parent: _GCPApi
        :param cache: The discovery document cache implementation. If set to
            None, an internal MemoryCache will be used.
        :type cache: Cache
        :return: The googleapiclient service or sub Resource.
        :rtype: _GCPApi
        """
        self._hub = hub
        self._parent = parent
        self._resources = {}
        self._api_resource = None
        if cache == None:
            self._cache = _GCPApi._DISCOVERY_CACHE
        else:
            self._cache = cache
        self.name = resource_type
        self._is_operation = (
            resource_type in ("zoneOperations", "regionOperations", "globalOperations")
            or self._parent
            and self._parent._is_operation
        )

    def get_dotted_notation(self) -> str:
        """Builds fully qualified 'dotted' path to self."""
        if self._parent is None:
            return self.name
        else:
            str = self._parent.get_dotted_notation()
            return f"{str}.{self.name}"

    def _build_service(self, ctx):
        """Return the top level gcp (client) service Resource.

        :param ctx: The Idem acct context (credentials) to use for
            the Google API calls.
        return: Google API Resource this instance wraps.
        :rtype: Resource
        """
        api_versions = ctx.acct.get("api_versions", {})
        api_version = api_versions.get(
            self.name, self._hub.tool.gcp.DEFAULT_VERSIONS[self.name]
        )

        ret = discovery.build(
            serviceName=self.name,
            version=api_version,
            credentials=ctx.acct["credentials"],
            cache=self._cache,
            discoveryServiceUrl=ctx.acct.get("discovery_service_url"),
        )
        self._api_resource = ret

        return ret

    def __getattr__(self, attr: str):
        """Gets and returns the named _GCPApi child.

        :type self: _GCPAPI
        :param attr: The name of the service or subresource, e.g.,
            'iam', or 'roles' or the like.
        :type attr: str
        :return: The googleapiclient (client) resource.
        :rtype: _GCPApi
        """
        if attr == "wait_on_operations":
            return self.wait_on_operations

        # If no existing resource exists, create it and attach it as a child
        try:
            ret = self._resources[attr]
        except KeyError:
            ret = _GCPApi(self._hub, attr, self)
            self._resources[attr] = ret

        return self._resources[attr]

    def build_call_resource(self, ctx, *args, **kwargs):
        """Recursively builds all _GCPApi Resources.

        This call walks a branch of a Google discovered API tree. For
        example, a call such as:

            my_gcpapi.iam.roles.list(ctx, *args, **kwargs)

        would result in a the creation (if not already created)
        the necessary _GCPApi wrappers for each of the dotted
        notation entries (indirectly via __getattr__). As well,
        for each of those wrappers, this call will create the
        Google API Resource that relates.

        :param ctx: The Idem acct context (credentials) to use for
            the Google API calls.
        :param args: A list of positional arguments to pass, as needed based
            on Google API signatures, each Resource to obtain subresources.
        type args: tuple
        :param kwargs: Named parameters as key/value pairs.
        :type kwargs: dictionary
        :return: The googleapiclient (client) resource.
        :rtype: _GCPApi
        """
        if self._parent is None:
            # Terminate the recursion at the top of the call tree:
            # the top of the tree is the cloud api object and has only
            # child callable services, therefore nothing to contribute.
            return None

        # build the parent api resource
        resource = self._parent.build_call_resource(ctx, *args, **kwargs)

        if resource is None:
            # This is a top level service (e.g., iam), return the result of
            # discovery build (i.e., the gcp api service client).
            if self._api_resource is None:
                self._api_resource = self._build_service(ctx)
            return self._api_resource
        else:
            # Subresource -- use the parent resource directly to avoid
            # another recursive buildout; it should be already created.
            resource = getattr(self._parent._api_resource, self.name)
            call_args = []
            call_kwargs = {}
            sig = signature(resource)
            for param in sig.parameters.values():
                if param.kind == param.POSITIONAL_ONLY:
                    call_args = args[call_args_index]
                    call_args_index + 1
                elif param.kind == param.POSITIONAL_OR_KEYWORD:
                    self._hub.log.debug("Error: Unsupported -- POSITIONAL_OR_KEYWORD")
                elif param.kind == param.KEYWORD_ONLY:
                    self._hub.log.debug("Error: Unsupported -- KEYWORD_ONLY")
                elif param.kind == param.VAR_POSITIONAL:
                    self._hub.log.debug("Error: Unsupported -- VAR_POSITIONAL")
                elif param.kind == param.VAR_KEYWORD:
                    call_kwargs = kwargs
            self._api_resource = resource(*call_args, **call_kwargs)
            return self._api_resource

    def wait_on_operations(self, wait: bool):
        """Set this object instance to wait for the completion, or related error, of all long running GCP API calls.

        :param wait: Whether to wait (True) or not (False).
        """
        _GCPApi.WAIT_ON_OPERATIONS = wait

    def _wait_on_operation(self, ctx, project, operation):
        """Waits for an operation to complete or error out.

        :type self: _GCPAPI
        :param ctx: The name of the service or subresource, e.g.,
            'iam', or 'roles' or the like.
        :param operation: The GCP API operation on which to wait.
        """
        if self.WAIT_ON_OPERATIONS:
            done = False
            kind = operation["kind"].split("#")[0]
            gcpapi = getattr(self._hub.tool.gcp.API, kind)
            call_args = []
            call_kwargs = {}
            call_kwargs["project"] = project
            call_kwargs["operation"] = operation["name"]
            while not done:
                # The wait calls below wait until the operation is DONE.
                # However, the GCP wait API makes no promises it will
                # truly wait until DONE, it is merely a 'best effort'
                # so multiple calls to wait may be required. Sleep one
                # second between to prevent potential call bursts.
                time.sleep(1)
                if "zone" in operation:
                    # The zone may get returned as a full URL such as
                    # https://www.googleapis.com/compute/v1/projects/project/zones/us-central1-a.
                    # In that case, parse it
                    path_split = urlparse(operation["zone"]).path.split("/")
                    last_path_element = path_split[len(path_split) - 1]
                    call_kwargs["zone"] = last_path_element
                    api = gcpapi.zoneOperations.wait
                elif "region" in operation:
                    # The region may get returned as a full URL. In that case, parse it
                    path_split = urlparse(operation["region"]).path.split("/")
                    last_path_element = path_split[len(path_split) - 1]
                    call_kwargs["region"] = last_path_element
                    api = gcpapi.regionOperations.wait
                else:
                    api = gcpapi.globalOperations.wait

                call = api.build_call_resource(ctx, *call_args, **call_kwargs)
                try:
                    operation = call.execute()
                    done = operation["status"] == "DONE"
                except timeout:
                    # This hits if the operation fails to finish
                    # though is still running. Other exceptions
                    # gather more deleterious failures.
                    done = False

        return operation

    async def __call__(self, ctx, *args, **kwargs) -> dict:
        """Calls the Google API associated with this instance (self).

        :param ctx: The Idem acct context (credentials) to use for
            the Google API calls.
        :param args: A list of positional arguments to pass, as needed based
            on Google API signatures, each Resource to obtain subresources.
        type args: tuple
        :param kwargs: Named parameters as key/value pairs. These will correspond
            to the Google Google API Python client library API specifics. `See, e.g.,
            <https://googleapis.github.io/google-api-python-client/docs/dyn/compute_v1.instances.html#insert>_`
            for an example of a large set of kwargs.
        :type kwargs: dictionary
        :return: The api client return value.
        """
        call = self.build_call_resource(ctx, *args, **kwargs)
        try:
            resp = call.execute()
            ret = resp
        except discovery.HttpError as httpError:
            if httpError.status_code == 404:
                raise idemexc.NotFoundError(str(httpError))
            else:
                raise httpError
        #     ret['error'] = str(httpError)
        #     ret['code'] = httpError.status_code
        # except discovery.InvalidJsonError as jsonError:
        #     ret['error'] = str(jsonError)
        #     ret['code'] = _GCPApi.E_JSONERROR
        # except discovery.MediaUploadSizeError as mediaError:
        #     ret['error'] = str(mediaError)
        #     ret['code'] = _GCPApi.E_MUTUALTLSCHANNELERROR
        # except discovery.UnacceptableMimeTypeError as mimeError:
        #     ret['error'] = str(mimeError)
        #     ret['code'] = _GCPApi.E_UNACCEPTABLEMIMETYPEERROR

        if "warnings" in resp:
            self._hub.log.warning(str(resp["warnings"]))

        if "error" in resp:
            raise idemexc.OperationError(resp["error"])

        if "kind" in resp and resp["kind"].endswith("#operation"):
            ret = self._wait_on_operation(ctx, kwargs.get("project", None), resp)

        return ret



"""Exec module for managing ServiceAccounts."""
from typing import Callable

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext

__func_alias__ = {"list_": "list"}


async def list_(hub, ctx, project: str = None):
    r"""Lists every ServiceAccount that belongs to a specific project.

    Args:
        project(str, Required):
            The resource name of the project associated with the service accounts.
    """
    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="iam.projects.service_account",
        method_name="list",
        method_params={"ctx": ctx, "name": f"projects/{project}"},
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def get(
    hub,
    ctx,
    project: str = None,
    unique_id: str = None,
    email: str = None,
    resource_id: str = None,
):
    r"""Returns the specified ServiceAccount resource.

    Args:
        project(str, Optional):
            Project ID for this request.
        unique_id(str, Optional):
            The unique, stable numeric ID for the service account.
        email(str, Optional):
            The email address of the service account.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    if unique_id or email:
        project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
        identifier = unique_id or email
        resource_id = f"projects/{project}/serviceAccounts/{identifier}"
    elif not resource_id:
        result["result"] = False
        result["comment"] = [
            f"gcp.iam.projects.service_account#get(): either resource_id or unique_id or email"
            f" should be specified."
        ]
        return result

    execution_context = ExecutionContext(
        resource_type="iam.projects.service_account",
        method_name="get",
        method_params={"ctx": ctx, "name": resource_id},
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def undelete(
    hub,
    ctx,
    unique_id: str,
    project: str = None,
):
    r"""Restores a deleted service account.

    It is not always possible to restore a deleted service account. Use this method only as a last resort.

    After you delete a service account, IAM permanently removes the service account 30 days later. There is no way to
    restore a deleted service account that has been permanently removed.

    The permission 'iam.serviceAccounts.undelete' is required for undeleting resources.

    Although the GCP documentation states that either email or unique_id can be used as resource name, using an email
    results in the following error: "The service account name must be in the following format:
    projects/{PROJECT_ID}/serviceAccounts/{ACCOUNT_UNIQUE_ID}".

    Args:
        unique_id(str):
            The unique, stable numeric ID for the service account.
        project(str, Optional):
            Project ID for this request.
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    if not unique_id:
        result["result"] = False
        result["comment"] = [
            f"gcp.iam.projects.service_account#undelete(): unique_id is required"
        ]
        return result

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
    resource_id = f"projects/{project}/serviceAccounts/{unique_id}"

    execution_context = ExecutionContext(
        resource_type="iam.projects.service_account",
        method_name="undelete",
        method_params={"ctx": ctx, "name": resource_id},
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def disable(
    hub,
    ctx,
    project: str = None,
    unique_id: str = None,
    email: str = None,
    resource_id: str = None,
):
    r"""Disables a service account immediately.

    If an application uses the service account to authenticate, that application can no longer call Google APIs or access Google Cloud resources. Existing access tokens for the service account are rejected, and requests for new access tokens will fail.

    To re-enable the service account, use service_account.enable. After you re-enable the service account, its existing access tokens will be accepted, and you can request new access tokens.

    To help avoid unplanned outages, we recommend that you disable the service account before you delete it. Use this method to disable the service account, then wait at least 24 hours and watch for unintended consequences.

    Args:
        project(str, Optional):
            Project ID for this request.
        unique_id(str, Optional):
            The unique, stable numeric ID for the service account.
        email(str, Optional):
            The email address of the service account.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.
    """

    def validate(get_ret):
        assert get_ret.get("disabled")

    return await _api_op(
        hub,
        ctx,
        "disable",
        validate,
        project=project,
        unique_id=unique_id,
        email=email,
        resource_id=resource_id,
    )


async def enable(
    hub,
    ctx,
    project: str = None,
    unique_id: str = None,
    email: str = None,
    resource_id: str = None,
):
    r"""Enables a service account that have previously been disabled.

     If the service account is already enabled, then this method has no effect.

     If the service account was disabled by other means — for example, if Google disabled the service account because it was compromised—you cannot use this method to enable the service account.

    Args:
        project(str, Optional):
            Project ID for this request.
        unique_id(str, Optional):
            The unique, stable numeric ID for the service account.
        email(str, Optional):
            The email address of the service account.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.
    """

    def validate(get_ret):
        assert "disabled" not in get_ret

    return await _api_op(
        hub,
        ctx,
        "enable",
        validate,
        project=project,
        unique_id=unique_id,
        email=email,
        resource_id=resource_id,
    )


async def _api_op(
    hub,
    ctx,
    api_method: str,
    validate: Callable,
    project: str = None,
    unique_id: str = None,
    email: str = None,
    resource_id: str = None,
):
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    if unique_id or email:
        project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
        identifier = unique_id or email
        resource_id = f"projects/{project}/serviceAccounts/{identifier}"
    elif not resource_id:
        result["result"] = False
        result["comment"] = [
            f"gcp.iam.projects.service_account#{api_method}(): "
            f"either resource_id or unique_id or email should be specified."
        ]
        return result

    async def validate_resource():
        """
        Validate subsequent get requests return expected output.
        """
        get_ret = await hub.tool.gcp.generate.generic_exec.execute(
            ExecutionContext(
                resource_type="iam.projects.service_account",
                method_name="get",
                method_params={"ctx": ctx, "name": resource_id},
            )
        )
        try:
            assert get_ret["result"]
            validate(get_ret["ret"])
        except AssertionError:
            # return dummy rerun_data to activate reconciliation
            result["rerun_data"] = {"api_method": api_method}
            result["comment"] += get_ret["comment"]
            result["result"] = False
            return result

        result["ret"] = get_ret["ret"]
        return result

    if ctx.get("rerun_data"):
        return await validate_resource()

    execution_context = ExecutionContext(
        resource_type="iam.projects.service_account",
        method_name=api_method,
        method_params={"ctx": ctx, "name": resource_id},
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    return await validate_resource()



"""State module for managing Buckets."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

RESOURCE_TYPE_SIMPLE = "storage.bucket"
RESOURCE_TYPE_FULL = f"gcp.{RESOURCE_TYPE_SIMPLE}"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    project: str = None,
    predefined_acl: str = None,
    predefined_default_object_acl: str = None,
    user_project: str = None,
    billing: make_dataclass(
        "BillingProperties",
        [("requester_pays", bool, field(default=None))],
    ) = None,
    cors: List[
        make_dataclass(
            "CorsProperties",
            [
                ("max_age_seconds", int, field(default=None)),
                ("method", List[str], field(default=None)),
                ("origin", List[str], field(default=None)),
                ("response_header", List[str], field(default=None)),
            ],
        )
    ] = None,
    custom_placement_config: make_dataclass(
        "CustomPlacementConfigProperties",
        [("data_locations", List[str], field(default=None))],
    ) = None,
    default_event_based_hold: bool = None,
    encryption: make_dataclass(
        "EncryptionProperties",
        [("default_kms_key_name", str, field(default=None))],
    ) = None,
    iam_configuration: make_dataclass(
        "IamConfigurationProperties",
        [
            (
                "bucket_policy_only",
                make_dataclass(
                    "BucketPolicyOnlyProperties",
                    [
                        ("enabled", bool, field(default=None)),
                        ("locked_time", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "uniform_bucket_level_access",
                make_dataclass(
                    "UniformBucketLevelAccessProperties",
                    [
                        ("enabled", bool, field(default=None)),
                        ("locked_time", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("public_access_prevention", str, field(default=None)),
        ],
    ) = None,
    labels: Dict[str, Any] = None,
    lifecycle: make_dataclass(
        "LifecycleProperties",
        [
            (
                "rule",
                List[
                    make_dataclass(
                        "RuleProperties",
                        [
                            (
                                "action",
                                make_dataclass(
                                    "ActionProperties",
                                    [
                                        ("storage_class", str, field(default=None)),
                                        ("type_", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "condition",
                                make_dataclass(
                                    "ConditionProperties",
                                    [
                                        ("age", int, field(default=None)),
                                        ("created_before", str, field(default=None)),
                                        (
                                            "custom_time_before",
                                            str,
                                            field(default=None),
                                        ),
                                        (
                                            "days_since_custom_time",
                                            int,
                                            field(default=None),
                                        ),
                                        (
                                            "days_since_noncurrent_time",
                                            int,
                                            field(default=None),
                                        ),
                                        ("is_live", bool, field(default=None)),
                                        ("matches_pattern", str, field(default=None)),
                                        (
                                            "matches_prefix",
                                            List[str],
                                            field(default=None),
                                        ),
                                        (
                                            "matches_suffix",
                                            List[str],
                                            field(default=None),
                                        ),
                                        (
                                            "matches_storage_class",
                                            List[str],
                                            field(default=None),
                                        ),
                                        (
                                            "noncurrent_time_before",
                                            str,
                                            field(default=None),
                                        ),
                                        (
                                            "num_newer_versions",
                                            int,
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    )
                ],
                field(default=None),
            )
        ],
    ) = None,
    autoclass: make_dataclass(
        "AutoclassProperties",
        [
            ("enabled", bool, field(default=None)),
        ],
    ) = None,
    location: str = None,
    logging: make_dataclass(
        "LoggingProperties",
        [
            ("log_bucket", str, field(default=None)),
            ("log_object_prefix", str, field(default=None)),
        ],
    ) = None,
    retention_policy: make_dataclass(
        "RetentionPolicyProperties",
        [
            ("retention_period", str, field(default=None)),
            ("is_locked", bool, field(default=None)),
        ],
    ) = None,
    rpo: str = None,
    storage_class: str = None,
    versioning: make_dataclass(
        "VersioningProperties",
        [("enabled", bool, field(default=None))],
    ) = None,
    website: make_dataclass(
        "WebsiteProperties",
        [
            ("main_page_suffix", str, field(default=None)),
            ("not_found_page", str, field(default=None)),
        ],
    ) = None,
    metageneration: str = None,
    if_metageneration_match: str = None,
    acl: List[
        make_dataclass(
            "BucketAccessControl",
            [
                ("entity", str, field(default=None)),
                ("role", str, field(default=None)),
            ],
        )
    ] = None,
    default_object_acl: List[
        make_dataclass(
            "ObjectAccessControl",
            [
                ("entity", str, field(default=None)),
                ("role", str, field(default=None)),
            ],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""Create or update a storage bucket resource. Changes to the bucket will be readable immediately after writing, but configuration changes may take time to propagate.

        Args:
            name(str): An Idem name of the resource.
            resource_id(str, Optional): An identifier of the resource in the provider. Defaults to None.
            project(str): A valid API project identifier.
            predefined_acl(str, optional): Apply a predefined set of access controls to this bucket. Defaults to None.
            predefined_default_object_acl(str, optional): Apply a predefined set of default object access controls to this bucket. Defaults to None.
            user_project(str, optional): The project to be billed for this request. Required for Requester Pays buckets. Defaults to None.
            billing(Dict[str, Any], optional): The bucket's billing configuration. Defaults to None.
                * requester_pays (bool, optional): When set to true, Requester Pays is enabled for this bucket.
            cors(List[Dict[str, Any]], optional): The bucket's Cross-Origin Resource Sharing (CORS) configuration. Defaults to None.
                * max_age_seconds (int, optional): The value, in seconds, to return in the  Access-Control-Max-Age header used in preflight responses.
                * method (List[str], optional): The list of HTTP methods on which to include CORS response headers, (GET, OPTIONS, POST, etc) Note: "*" is permitted in the list of methods, and means "any method".
                * origin (List[str], optional): The list of Origins eligible to receive CORS response headers. Note: "*" is permitted in the list of origins, and means "any Origin".
                * response_header (List[str], optional): The list of HTTP headers other than the simple response headers to give permission for the user-agent to share across domains.
            custom_placement_config(Dict[str, Any], optional): The bucket's custom placement configuration for Custom Dual Regions. Defaults to None.
                * data_locations (List[str], optional): The list of regional locations in which data is placed.
            default_event_based_hold(bool, optional): The default value for event-based hold on newly created objects in this bucket. Event-based hold is a way to retain objects indefinitely until an event occurs, signified by the hold's release. After being released, such objects will be subject to bucket-level retention (if any). One sample use case of this flag is for banks to hold loan documents for at least 3 years after loan is paid in full. Here, bucket-level retention is 3 years and the event is loan being paid in full. In this example, these objects will be held intact for any number of years until the event has occurred (event-based hold on the object is released) and then 3 more years after that. That means retention duration of the objects begins from the moment event-based hold transitioned from true to false. Objects under event-based hold cannot be deleted, overwritten or archived until the hold is removed. Defaults to None.
            encryption(Dict[str, Any], optional): Encryption configuration for a bucket. Defaults to None.
                * default_kms_key_name (str, optional): A Cloud KMS key that will be used to encrypt objects inserted into this bucket, if no encryption method is specified.
            iam_configuration(Dict[str, Any], optional): The bucket's IAM configuration. Defaults to None.
                * bucket_policy_only (Dict[str, Any], optional): The bucket's uniform bucket-level access configuration. The feature was formerly known as Bucket Policy Only. For backward compatibility, this field will be populated with identical information as the uniformBucketLevelAccess field. We recommend using the uniformBucketLevelAccess field to enable and disable the feature.
                    * enabled (bool, optional): If set, access is controlled only by bucket-level or above IAM policies.
                    * locked_time (str, optional): The deadline for changing iamConfiguration.bucketPolicyOnly.enabled from true to false in RFC 3339 format. iamConfiguration.bucketPolicyOnly.enabled may be changed from true to false until the locked time, after which the field is immutable.
                * uniform_bucket_level_access (Dict[str, Any], optional): The bucket's uniform bucket-level access configuration.
                    * enabled (bool, optional): If set, access is controlled only by bucket-level or above IAM policies.
                    * locked_time (str, optional): The deadline for changing iamConfiguration.uniformBucketLevelAccess.enabled from true to false in RFC 3339  format. iamConfiguration.uniformBucketLevelAccess.enabled may be changed from true to false until the locked time, after which the field is immutable.
                * public_access_prevention (str, optional): The bucket's Public Access Prevention configuration. Currently, 'inherited' and 'enforced' are supported.
            labels(Dict[str, Any], optional): User-provided labels, in key/value pairs. Defaults to None.
            lifecycle(Dict[str, Any], optional): The bucket's lifecycle configuration. See lifecycle management for more information. Defaults to None.
                * rule (List[Dict[str, Any]], optional): A lifecycle management rule, which is made of an action to take and the condition(s) under which the action will be taken.
                    * action (Dict[str, Any], optional): The action to take.
                        * storage_class (str, optional): Target storage class. Required iff the type of the action is SetStorageClass.
                        * type_ (str, optional): Type of the action. Currently, only Delete, SetStorageClass, and AbortIncompleteMultipartUpload are supported.
                    * condition (Dict[str, Any], optional): The condition(s) under which the action will be taken.
                        * age (int, optional): Age of an object (in days). This condition is satisfied when an object reaches the specified age.
                        * created_before (str, optional): A date in RFC 3339 format with only the date part (for instance, "2013-01-15"). This condition is satisfied when an object is created before midnight of the specified date in UTC.
                        * custom_time_before (str, optional): A date in RFC 3339 format with only the date part (for instance, "2013-01-15"). This condition is satisfied when the custom time on an object is before this date in UTC.
                        * days_since_custom_time (int, optional): Number of days elapsed since the user-specified timestamp set on an object. The condition is satisfied if the days elapsed is at least this number. If no custom timestamp is specified on an object, the condition does not apply.
                        * days_since_noncurrent_time (int, optional): Number of days elapsed since the noncurrent timestamp of an object. The condition is satisfied if the days elapsed is at least this number. This condition is relevant only for versioned objects. The value of the field must be a nonnegative integer. If it's zero, the object version will become eligible for Lifecycle action as soon as it becomes noncurrent.
                        * is_live (bool, optional): Relevant only for versioned objects. If the value is true, this condition matches live objects; if the value is false, it matches archived objects.
                        * matches_pattern (str, optional): A regular expression that satisfies the RE2 syntax. This condition is satisfied when the name of the object matches the RE2 pattern. Note: This feature is currently in the "Early Access" launch stage and is only available to a whitelisted set of users; that means that this feature may be changed in backward-incompatible ways and that it is not guaranteed to be released.
                        * matches_prefix (List[str], optional): List of object name prefixes. This condition will be satisfied when at least one of the prefixes exactly matches the beginning of the object name.
                        * matches_suffix (List[str], optional): List of object name suffixes. This condition will be satisfied when at least one of the suffixes exactly matches the end of the object name.
                        * matches_storage_class (List[str], optional): Objects having any of the storage classes specified by this condition will be matched. Values include MULTI_REGIONAL, REGIONAL, NEARLINE, COLDLINE, ARCHIVE, STANDARD, and DURABLE_REDUCED_AVAILABILITY.
                        * noncurrent_time_before (str, optional): A date in RFC 3339 format with only the date part (for instance, "2013-01-15"). This condition is satisfied when the noncurrent time on an object is before this date in UTC. This condition is relevant only for versioned objects.
                        * num_newer_versions (int, optional): Relevant only for versioned objects. If the value is N, this condition is satisfied when there are at least N versions (including the live version) newer than this version of the object.
            autoclass(Dict[str, Any], optional): The bucket's Autoclass configuration. Defaults to None.
                * enabled (bool, optional): Whether or not Autoclass is enabled on this bucket
                * toggle_time (str, optional): A date and time in RFC 3339 format representing the instant at which "enabled" was last toggled.
            location(str, optional): The location of the bucket. Object data for objects in the bucket resides in physical storage within this region. Defaults to US. See the developer's guide for the authoritative list. Defaults to None.
            logging(Dict[str, Any], optional): The bucket's logging configuration, which defines the destination bucket and optional name prefix for the current bucket's logs. Defaults to None.
                * log_bucket (str, optional): The destination bucket where the current bucket's logs should be placed.
                * log_object_prefix (str, optional): A prefix for log object names.
            retention_policy(Dict[str, Any], optional): The bucket's retention policy. The retention policy enforces a minimum retention time for all objects contained in the bucket, based on their creation time. Any attempt to overwrite or delete objects younger than the retention period will result in a PERMISSION_DENIED error. An unlocked retention policy can be modified or removed from the bucket via a storage.buckets.update operation. A locked retention policy cannot be removed or shortened in duration for the lifetime of the bucket. Attempting to remove or decrease period of a locked retention policy will result in a PERMISSION_DENIED error. Defaults to None.
                * is_locked (bool, optional): Locks the retention policy. Once locked, an object retention policy cannot be modified.
                * retention_period (str, optional): The duration in seconds that objects need to be retained. Retention duration must be greater than zero and less than 100 years. Note that enforcement of retention periods less than a day is not guaranteed. Such periods should only be used for testing purposes.
            rpo(str, optional): The Recovery Point Objective (RPO) of this bucket. Set to ASYNC_TURBO to turn on Turbo Replication on a bucket. Defaults to None.
            storage_class(str, optional): The bucket's default storage class, used whenever no storageClass is specified for a newly-created object. This defines how objects in the bucket are stored and determines the SLA and the cost of storage. Values include MULTI_REGIONAL, REGIONAL, STANDARD, NEARLINE, COLDLINE, ARCHIVE, and DURABLE_REDUCED_AVAILABILITY. If this value is not specified when the bucket is created, it will default to STANDARD. For more information, see storage classes. Defaults to None.
            versioning(Dict[str, Any], optional): The bucket's versioning configuration. Defaults to None.
                * enabled (bool, optional): While set to true, versioning is fully enabled for this bucket.
            website(Dict[str, Any], optional): The bucket's website configuration, controlling how the service behaves when accessing bucket contents as a web site. See the Static Website Examples for more information. Defaults to None.
                * main_page_suffix (str, optional): If the requested object path is missing, the service will ensure the path has a trailing '/', append this suffix, and attempt to retrieve the resulting object. This allows the creation of index.html objects to represent directory pages.
                * not_found_page (str, optional): If the requested object path is missing, and any mainPageSuffix object is missing, if applicable, the service will return the named object from this bucket as the content for a 404 Not Found result.
            metageneration(str, optional): The metadata generation of this bucket. Read-only property
            if_metageneration_match(str, optional): Makes the operation conditional on whether bucket's current metageneration matches the given value. Currently used only when locking retention policy
            acl(list[Dict[str, Any]], Optional): Access controls on the bucket. Defaults to None.
                * entity (str, Optional): The entity holding the permission, in one of the following forms:
                    - user-userId
                    - user-email
                    - group-groupId
                    - group-email
                    - domain-domain
                    - project-team-projectId
                    - allUsers
                    - allAuthenticatedUsers Examples:
                    - The user liz@example.com would be user-liz@example.com.
                    - The group example@googlegroups.com would be group-example@googlegroups.com.
                    - To refer to all members of the Google Apps for Business domain example.com, the entity would be domain-example.com.
                * role (str, Optional): The access permission for the entity.
            default_object_acl(list[Dict[str, Any]], Optional): Default access controls to apply to new objects when no ACL is provided. Defaults to None.
                * entity (str, Optional): The entity holding the permission, in one of the following forms:
                    - user-userId
                    - user-email
                    - group-groupId
                    - group-email
                    - domain-domain
                    - project-team-projectId
                    - allUsers
                    - allAuthenticatedUsers Examples:
                    - The user liz@example.com would be user-liz@example.com.
                    - The group example@googlegroups.com would be group-example@googlegroups.com.
                    - To refer to all members of the Google Apps for Business domain example.com, the entity would be domain-example.com.
                * role (str, Optional): The access permission for the entity.
    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              gcp.storage.bucket.present:
                - name: value
                - project: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    resource_type_camel = hub.tool.gcp.case.camel(RESOURCE_TYPE_FULL.split(".")[-1])

    if hub.tool.gcp.resource_prop_utils.properties_mismatch_resource_id(
        RESOURCE_TYPE_SIMPLE, resource_id, {**locals(), resource_type_camel: name}
    ):
        result["comment"].append(
            hub.tool.gcp.comment_utils.properties_mismatch_resource_id_comment(
                RESOURCE_TYPE_FULL, name
            )
        )
    if resource_id:
        old_get_ret = await hub.exec.gcp.storage.bucket.get(
            ctx, resource_id=resource_id, projection="full", user_project=user_project
        )

        if not old_get_ret["result"] or not old_get_ret["ret"]:
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.resource_not_found_comment(
                    RESOURCE_TYPE_FULL, resource_id
                )
            )
            result["comment"].extend(old_get_ret["comment"])
            return result

        # if resource_id is provided explicitly or by ESM, we want update
        result["old_state"] = old_get_ret["ret"]
    elif not hub.OPT.idem.get("get_resource_only_with_resource_id", False):
        # attempt to discover existing resource which is not tracked by idem
        resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
            RESOURCE_TYPE_SIMPLE, {**locals(), "bucket": name}
        )

        old_get_ret = await hub.exec.gcp.storage.bucket.get(
            ctx, resource_id=resource_id, projection="full", user_project=user_project
        )

        if not old_get_ret["result"]:
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.resource_not_found_comment(
                    RESOURCE_TYPE_FULL, resource_id
                )
            )
            result["comment"].extend(old_get_ret["comment"])
            return result

        # resource we try to discover exists
        if old_get_ret["ret"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.resource_discovered_comment(
                    RESOURCE_TYPE_FULL, resource_id
                )
            )
            result["old_state"] = old_get_ret["ret"]

    old = result["old_state"]

    resource_body = {
        "name": name,
        "autoclass": autoclass,
        "billing": billing,
        "cors": cors,
        "custom_placement_config": custom_placement_config,
        "default_event_based_hold": default_event_based_hold,
        "encryption": encryption,
        "iam_configuration": iam_configuration,
        "labels": labels,
        "lifecycle": lifecycle,
        "location": location,
        "logging": logging,
        "retention_policy": retention_policy,
        "rpo": rpo,
        "storage_class": storage_class,
        "versioning": versioning,
        "website": website,
        "metageneration": metageneration,
        "acl": acl,
        "default_object_acl": default_object_acl,
    }

    plan_state = {
        "resource_id": resource_id,
        **resource_body,
    }

    plan_state = {k: v for (k, v) in plan_state.items() if v is not None}

    if old:
        result["new_state"] = old

        changes = hub.tool.gcp.utils.compare_states(
            old, plan_state, RESOURCE_TYPE_SIMPLE
        )

        if not changes:
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(RESOURCE_TYPE_FULL, name)
            )

            return result

        changed_non_updatable_properties = (
            hub.tool.gcp.resource_prop_utils.get_changed_non_updatable_properties(
                RESOURCE_TYPE_SIMPLE, changes
            )
        )

        if changed_non_updatable_properties:
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                    RESOURCE_TYPE_FULL, name, changed_non_updatable_properties
                )
            )
            return result

        # is_locked property is not updateable through the regular patch operation
        new_is_locked = (plan_state.get("retention_policy") or {}).pop(
            "is_locked", None
        )

        if changes.get("relevant_changes", {}) != {
            "root['retention_policy']['is_locked']"
        }:
            if ctx["test"]:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.would_update_comment(
                        RESOURCE_TYPE_FULL, name
                    )
                )
                result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                    {**old, **plan_state}
                )
            else:
                # Perform update
                update_ret = await hub.exec.gcp_api.client.storage.bucket.patch(
                    hub,
                    ctx,
                    name=name,
                    resource_id=resource_id,
                    body=plan_state,
                    userProject=user_project,
                )

                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] += update_ret["comment"]
                    return result

                result["comment"].append(
                    hub.tool.gcp.comment_utils.update_comment(RESOURCE_TYPE_FULL, name)
                )

                if not if_metageneration_match:
                    if_metageneration_match = update_ret.get("ret", {}).get(
                        "metageneration"
                    )

        # lock retention policy after update has succeeded to ensure that the retention period specified
        # has been applied
        old_is_locked = (old.get("retention_policy") or {}).get("is_locked")

        if not if_metageneration_match:
            if_metageneration_match = old.get("metageneration")
    else:
        result["old_state"] = None
        old_is_locked = False
        new_is_locked = (resource_body.get("retention_policy") or {}).pop(
            "is_locked", None
        )
        (resource_body.get("retention_policy") or {}).pop("effective_time", None)

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    RESOURCE_TYPE_FULL, name
                )
            )
            hub.tool.gcp.resource_prop_utils.populate_resource_with_assumed_values(
                plan_state,
                RESOURCE_TYPE_SIMPLE,
                convert_to_present=True,
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                plan_state
            )

            if not if_metageneration_match:
                if_metageneration_match = "mock_value"
        else:
            project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

            # Create
            create_ret = await hub.exec.gcp_api.client.storage.bucket.insert(
                ctx,
                name=name,
                project=project,
                userProject=user_project,
                predefined_acl=predefined_acl,
                predefined_default_object_acl=predefined_default_object_acl,
                projection="full",
                body=resource_body,
            )

            if not create_ret["result"]:
                result["result"] = False
                if not create_ret["comment"]:
                    return result
                if next(
                    (
                        comment
                        for comment in create_ret["comment"]
                        if "HttpError 409" in comment
                    ),
                    None,
                ):
                    result["comment"].append(
                        hub.tool.gcp.comment_utils.already_exists_comment(
                            RESOURCE_TYPE_FULL, name
                        )
                    )
                    return result
                else:
                    result["comment"].extend(create_ret["comment"])
                return result

            result["comment"].append(
                hub.tool.gcp.comment_utils.create_comment(RESOURCE_TYPE_FULL, name)
            )

            resource_id = create_ret.get("ret", {}).get("resource_id")

            if not if_metageneration_match:
                if_metageneration_match = create_ret.get("ret", {}).get(
                    "metageneration"
                )

    enforce_retention_policy_result = (
        await hub.tool.gcp.storage.bucket.enforce_retention_policy(
            ctx,
            result,
            old_is_locked,
            new_is_locked,
            if_metageneration_match,
            user_project,
            resource_id,
        )
    )

    if not ctx["test"]:
        await _assign_new_state_from_get(hub, ctx, user_project, result, resource_id)

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    user_project: str = None,
):
    r"""Permanently deletes an empty bucket.

    Args:
        name(str):
            The name of the resource

        resource_id(str, Optional):
            The resource_id of the resource

        user_project(str, Optional):
            The project to be billed for this request. Required for Requester Pays buckets.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        resource_is_absent:
          gcp.storage.bucket.absent

    """
    result = {
        "result": True,
        "old_state": ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "comment": [],
    }

    if not resource_id and not hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    ):
        resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
            RESOURCE_TYPE_SIMPLE, {**locals(), "bucket": name}
        )

    if not resource_id:
        # we don't have enough information to know what to delete
        # TODO: should this be an error? The SLS inputs are invalid
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(RESOURCE_TYPE_FULL, name)
        )
        return result

    get_ret = await hub.exec.gcp.storage.bucket.get(
        ctx, resource_id=resource_id, projection="full", user_project=user_project
    )

    if not get_ret["result"]:
        result["result"] = False
        result["comment"].append(
            hub.tool.gcp.comment_utils.resource_not_found_comment(
                RESOURCE_TYPE_FULL, resource_id
            )
        )
        result["comment"].extend(get_ret["comment"])
        return result

    if not get_ret["ret"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(RESOURCE_TYPE_FULL, name)
        )
        return result

    result["old_state"] = get_ret["ret"]

    if ctx["test"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(RESOURCE_TYPE_FULL, name)
        )
        return result

    # Delete the existing bucket (async call).
    del_ret = await hub.exec.gcp_api.client.storage.bucket.delete(
        ctx, resource_id=resource_id, userProject=user_project
    )

    if not del_ret["result"]:
        result["result"] = False
        result["comment"].extend(del_ret["comment"])
        return result

    result["comment"].append(
        hub.tool.gcp.comment_utils.delete_comment(RESOURCE_TYPE_FULL, name)
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of buckets.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.storage.bucket
    """
    result = {}

    # TODO: Pagination
    describe_ret = await hub.exec.gcp.storage.bucket.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe buckets {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.storage.bucket.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)


async def _assign_new_state_from_get(hub, ctx, user_project, result, resource_id):
    get_ret = await hub.exec.gcp.storage.bucket.get(
        ctx, resource_id=resource_id, projection="full", user_project=user_project
    )

    if not get_ret["result"] or not get_ret["ret"]:
        result["result"] = False
        result["comment"].append(
            hub.tool.gcp.comment_utils.resource_not_found_comment(
                RESOURCE_TYPE_FULL, resource_id
            )
        )
        result["comment"].extend(get_ret["comment"])
        return

    result["new_state"] = get_ret["ret"]



import dataclasses
import json
import typing
import uuid
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List

import typing_extensions
from cloudspec import CloudSpecParam


@dataclass
class ResourceProperty:
    type_: str = (
        None  # "string" | "boolean" | "integer" | "number" | "object" | "array" | null
    )
    description: str = None
    _ref: str = None  # $ref
    items: "ResourceProperty" = None  # used with properties of type "array"
    properties: Dict[
        str, "ResourceProperty"
    ] = None  # used with properties of type "object" for representing complex object
    enum: List[str] = None
    enum_descriptions: List[str] = None

    format: str = None  # "byte" | "double" | "float" | "int32" | "int64" | "uint32" | "uint64" | null
    # can be used to detail what type the string holds
    additional_properties: Dict[
        str, str
    ] = None  # can be used for more fine-grained typing?
    annotations: Dict[
        str, Any
    ] = None  # can be used to see if it is required for methods


@dataclass
class ResourceSchema:
    """Class representing GCP resource schema as returned by discovery API
    See: https://www.googleapis.com/discovery/v1/apis/compute/v1/rest
    """

    id_: str = None  # ref name
    description: str = None
    properties: Dict[str, ResourceProperty] = None


def parse_file(hub, path: str = "") -> Dict[str, ResourceSchema]:
    with open(path) as json_schema_file:
        schemas = json.load(json_schema_file)
    return hub.autogen.gcp.schema_parser.parse_json(schemas["schemas"])


def parse_json(hub, schemas: Dict[str, Any]) -> Dict[str, ResourceSchema]:
    schemas = {
        resource_name: _dataclass_from_dict(
            hub.autogen.gcp.schema_parser.sanitize_schema(resource_schema),
            ResourceSchema,
        )
        for (resource_name, resource_schema) in schemas.items()
    }
    return schemas


def parse_schemas(hub, schemas: Dict[str, ResourceSchema]) -> Dict[str, CloudSpecParam]:
    parsed_schemas = dict()
    for resource_name in schemas:
        hub.autogen.gcp.schema_parser.parse_schema(
            resource_name, schemas, parsed_schemas
        )
    return parsed_schemas


def parse_schema(
    hub,
    resource_name: str,
    schemas: Dict[str, ResourceSchema],
    parsed_schemas: Dict[str, CloudSpecParam],
):
    """Parses the schema for resource_name and all its dependencies, writes results to parsed_schemas"""
    resource_schema = schemas.get(resource_name)
    result = CloudSpecParam(
        name="unknown",
        required=False,
        target="kwargs",
        target_type="mapping",
        param_type="{}",
    )
    if resource_schema.description:
        result["doc"] = resource_schema.description

    member = {"name": resource_name}
    properties = resource_schema.properties
    member["params"] = {
        property_name: hub.autogen.gcp.schema_parser.parse_resource_prop(
            resource_prop, schemas, parsed_schemas, property_name
        )
        for (property_name, resource_prop) in properties.items()
    }
    result["member"] = member
    parsed_schemas[resource_name] = result
    return result


def parse_resource_prop(
    hub,
    resource_prop: ResourceProperty,
    schemas: Dict[str, ResourceSchema],
    parsed_schemas: Dict[str, CloudSpecParam],
    name: str = None,
) -> CloudSpecParam:
    if not name:
        name = "unknown"
    # name is known only in the wrapper context - enclosing type field name or top level parameter name
    result = CloudSpecParam(
        name=name, required=False, target="kwargs", target_type="mapping", doc=""
    )

    if resource_prop.description:
        result["doc"] = resource_prop.description

    if resource_prop.enum:
        enum_doc = hub.autogen.gcp.schema_parser.generate_enum_doc(
            resource_prop.enum, resource_prop.enum_descriptions
        )
        result["doc"] = hub.autogen.gcp.schema_parser.join_docs(result["doc"], enum_doc)

    if resource_prop.type_ and resource_prop.type_ not in {"object", "array"}:
        # TODO: see if we use different type names than returned by the API
        result["param_type"] = hub.autogen.gcp.schema_parser.convert_primitive_type(
            resource_prop.type_
        )

    elif resource_prop.type_ == "array":
        if resource_prop.items:
            parsed_member_cloudspec = hub.autogen.gcp.schema_parser.parse_resource_prop(
                resource_prop.items, schemas, parsed_schemas
            )
            if (
                parsed_member_cloudspec.param_type == "{}"
                and parsed_member_cloudspec.member
            ):
                result["param_type"] = "List[{}]"
                result["member"] = parsed_member_cloudspec.member
            else:
                result["param_type"] = f"List[{parsed_member_cloudspec.param_type}]"
                result["member"] = parsed_member_cloudspec.member
            # member_name = resource_prop.items._ref
            # if not member_name:
            #     member_name = "properties"
            # result["member"] = {
            #     "name": member_name,
            #     "params": parsed_member_cloudspec
            # }
        else:
            result["param_type"] = "List[Any]"

    elif resource_prop.type_ == "object" and resource_prop.properties:
        result["param_type"] = "{}"
        # TODO: cover cases where object is a ref (has additional_properties._ref)
        member = {"name": _generate_random_classname()}
        properties = resource_prop.properties
        member["params"] = {
            property_name: hub.autogen.gcp.schema_parser.parse_resource_prop(
                resource_prop, schemas, parsed_schemas, property_name
            )
            for (property_name, resource_prop) in properties.items()
        }
        result["member"] = member

    elif resource_prop.type_ == "object":
        result["param_type"] = "Dict[str, Any]"

    else:
        result["param_type"] = "Any"

    if resource_prop._ref:
        resource_name = resource_prop._ref
        if resource_name in parsed_schemas:
            parsed_schema = parsed_schemas[resource_name]
        else:
            parsed_schema = hub.autogen.gcp.schema_parser.parse_schema(
                resource_name, schemas, parsed_schemas
            )

        result["param_type"] = parsed_schema["param_type"]
        result["member"] = parsed_schema["member"]

        if parsed_schema.doc:
            result["doc"] = hub.autogen.gcp.schema_parser.join_docs(
                result["doc"], resource_name + ": " + parsed_schema.doc
            )

        # todo: add doc here

    return result


def convert_primitive_type(hub, gcp_type: str) -> str:
    if gcp_type == "boolean":
        return "bool"
    elif gcp_type == "string":
        return "str"
    elif gcp_type == "integer":
        return "int"
    elif gcp_type == "number":
        return "float"
    return "Any"


def generate_enum_doc(hub, enum_list: List[str], enum_descriptions: List[str]) -> str:
    if not enum_descriptions or len(enum_descriptions) != len(enum_list):
        #  descriptions are not as expected
        enum_descriptions = ["" for enum_value in enum_list]

    zipped = zip(enum_list, enum_descriptions)
    quoted_values = [
        f'"{enum_value}" - {enum_desc}' if enum_desc else f'"{enum_value}"'
        for (enum_value, enum_desc) in zipped
    ]
    return "Enum type. Allowed values:\n" + "\n".join(quoted_values)


def join_docs(hub, doc1: str, doc2: str) -> str:
    # removing empty docs
    docs = [doc for doc in [doc1, doc2] if doc]
    return "\n".join(docs)


def _dataclass_from_dict(d: Dict[str, Any], klass):
    field_types = {
        class_field.name: class_field.type for class_field in dataclasses.fields(klass)
    }
    raw_field_values = {
        dict_key: dict_value
        for (dict_key, dict_value) in d.items()
        if dict_key in field_types.keys()
    }
    return klass(
        **{
            field_name: _parse_raw_field_value(
                raw_field_value, field_types.get(field_name)
            )
            for (field_name, raw_field_value) in raw_field_values.items()
        }
    )


def _parse_raw_field_value(raw_field_value: Any, expected_type) -> Any:
    if dataclasses.is_dataclass(expected_type):
        return _dataclass_from_dict(raw_field_value, expected_type)
    if (
        expected_type == typing.ForwardRef("ResourceProperty")
        or expected_type == "ResourceProperty"
    ):
        return _dataclass_from_dict(raw_field_value, ResourceProperty)
    if (
        isinstance(raw_field_value, list)
        and typing_extensions.get_origin(expected_type) == list
    ):
        return [
            _parse_raw_field_value(
                list_value, typing_extensions.get_args(expected_type)[0]
            )
            for list_value in raw_field_value
        ]
    if (
        isinstance(raw_field_value, dict)
        and typing_extensions.get_origin(expected_type) == dict
    ):
        return {
            key: _parse_raw_field_value(
                value, typing_extensions.get_args(expected_type)[1]
            )
            for (key, value) in raw_field_value.items()
        }
    return raw_field_value


def sanitize_schema(hub, schema: Any) -> Any:
    if isinstance(schema, dict):
        return {
            hub.autogen.gcp.schema_parser.sanitize_key(
                key
            ): hub.autogen.gcp.schema_parser.sanitize_schema(value)
            for (key, value) in schema.items()
        }
    if isinstance(schema, list):
        return [hub.autogen.gcp.schema_parser.sanitize_schema(item) for item in schema]
    return schema


def sanitize_key(hub, key: str) -> str:
    return hub.tool.gcp.case.unclash(hub.tool.gcp.case.snake(key))


def _generate_random_classname() -> str:
    return "Class" + str(uuid.uuid4()).replace("-", "")



"""Utility functions for crypto key version resources."""
import copy
import os
from typing import Any
from typing import Dict

from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import keywrap
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding


def to_state(hub, state: Dict[str, Any]) -> Dict[str, Any]:
    """`state` is reserved in Idem and can be a parameter name in `present`.

    This method returns a copy of the original new/old state to get rid of NamespacedDict
    and replaces `state` with `key_state`. GCP `name` is translated to Idem `resource_id`.

    Args:
        state: new_state or old_state formatted variable

    Returns:
        Dict[str, Any]
    """
    result = copy.copy(state)
    if "state" in result:
        key_state = result["state"]
        del result["state"]
        result["key_state"] = key_state

    if "resource_id" in result:
        els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
            "cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions",
            result["resource_id"],
        )
        result["crypto_key_version_id"] = els["crypto_key_version_id"]
    return result


def wrap_key(hub, formatted_key: bytes, import_job_pub_key: str) -> bytes:
    """
    Generates and imports local key material to Cloud KMS.

    Args:
        formatted_key (bytes): Key material.
        import_job_pub_key (str): PEM encoded import job public key.

    Returns:
        bytes
    """
    # Generate a temporary 32-byte key for AES-KWP and wrap the key material.
    kwp_key = os.urandom(32)
    wrapped_target_key = keywrap.aes_key_wrap_with_padding(
        kwp_key, formatted_key, backends.default_backend()
    )

    # Retrieve the public key from the import job.
    import_job_pub = serialization.load_pem_public_key(
        bytes(import_job_pub_key, "UTF-8"), backends.default_backend()
    )

    # Wrap the KWP key using the import job key.
    wrapped_kwp_key = import_job_pub.encrypt(
        kwp_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )
    return wrapped_kwp_key + wrapped_target_key


async def get_import_key_if_eligible(
    hub,
    ctx,
    has_key_material: bool,
    has_algorithm: str,
    import_job: str,
    crypto_key_version_state: str,
) -> Dict[str, Any]:
    r"""Retrieves public key from an import job if key_material is provided.

     This method checks that import job is ACTIVE and `CryptoKeyVersion`_ be in `DESTROYED`_ or `IMPORT_FAILED`_ state.
     The key material and algorithm must match the previous `CryptoKeyVersion`_ exactly if the `CryptoKeyVersion`_ has
     ever contained key material.

    .. _CryptoKeyVersion: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion
    .. _DESTROYED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.DESTROYED
    .. _IMPORT_FAILED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.IMPORT_FAILED

    Args:
        has_key_material(bool): Required. Key material for import is available.
        has_algorithm(bool): Required but may evaluate to None. Algorithm is provided.
        import_job(str): Required but may evaluate to None. Import job Idem resource_id.
        crypto_key_version_state(str):
            Required but may evaluate to None. State of the crypto_key_version if importing into old instance.

    Returns:
        Dict[str, Any]
    """
    result = {"result": True, "comment": [], "ret": None}
    if not has_key_material:
        return result
    if not has_algorithm or not import_job:
        result["result"] = False
        result["comment"].append(
            f"Import will be attempted because key_material was provided but either algorithm or import_job is not specified."
        )
        return result
    if crypto_key_version_state and crypto_key_version_state not in [
        "DESTROYED",
        "IMPORT_FAILED",
    ]:
        result["result"] = False
        result["comment"].append(
            f"Reimport will be attempted because key_material was provided but crypto_key_version is not in the proper state."
        )
        return result

    import_job_ret = await hub.exec.gcp.cloudkms.import_job.get(
        ctx, resource_id=import_job
    )

    if not import_job_ret["result"] or not import_job_ret["ret"]:
        result["result"] = False
        result["comment"] += import_job_ret["comment"]
        return result

    import_job_resource = import_job_ret["ret"]

    if import_job_resource["state"] != "ACTIVE":
        result["result"] = False
        result["comment"] += (
            f"Import job {import_job} should be in state ACTIVE but current state is {import_job_resource['state']}",
        )
        return result

    result["ret"] = import_job_resource["public_key"]["pem"]
    return result



"""Exec module for managing Node Templates."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def list_(
    hub,
    ctx,
    project: str = None,
    region: str = None,
    filter_: str = None,
    max_results: int = 500,
    order_by: str = None,
    page_token: str = None,
    return_partial_success: bool = False,
) -> Dict[str, Any]:
    r"""Retrieves a list of node templates available to the specified project.

    Args:
        project(str, Optional):
            Project ID for this request.

        region(str, Optional):
            The name of the region for this request.

        filter_(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`.

        max_results(int, Optional):
            The maximum number of results per page that should be returned. If the number of available results is larger than `maxResults`, Compute Engine returns a `nextPageToken` that can be used to get the next page of results in subsequent list requests. Acceptable values are `0` to `500`, inclusive. (Default: `500`)

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

        page_token(str, Optional):
            Specifies a page token to use. Set `pageToken` to the `nextPageToken` returned by a previous list request to get the next page of results.

        return_partial_success(bool, Optional):
            Opt-in for partial success behavior which provides partial results in case of failure. The default value is false.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.node_template.list
              - kwargs:
                  project: project-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if region:
        ret = await hub.exec.gcp_api.client.compute.node_template.list(
            ctx,
            project=project,
            region=region,
            filter=filter_,
            maxResults=max_results,
            orderBy=order_by,
            pageToken=page_token,
            returnPartialSuccess=return_partial_success,
        )
    else:
        ret = await hub.exec.gcp_api.client.compute.node_template.aggregatedList(
            ctx,
            project=project,
            filter=filter_,
            maxResults=max_results,
            orderBy=order_by,
            pageToken=page_token,
            returnPartialSuccess=return_partial_success,
        )

    if not ret["result"]:
        result["comment"] += ret["comment"]
        result["result"] = False
        return result

    result["ret"] = ret["ret"].get("items", [])
    return result


async def get(
    hub,
    ctx,
    resource_id: str = None,
    region: str = None,
    name: str = None,
    project: str = None,
) -> Dict[str, Any]:
    r"""Returns the specified node template. Gets a list of available node templates by making a list() request.

    Args:

        resource_id(str, Optional):
            An identifier of the resource in the provider.

        region(str, Optional):
            The name of the region for this request.

        name(str, Optional):
            Name of the node template to return.

        project(str, Optional):
            Project ID for this request.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.node_template.get
              - kwargs:
                  project: project-name
                  region: region
                  name: node-template-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if resource_id:
        ret = await hub.exec.gcp_api.client.compute.node_template.get(
            ctx,
            resource_id=resource_id,
        )
    elif project and region and name:
        ret = await hub.exec.gcp_api.client.compute.node_template.get(
            ctx, project=project, region=region, nodeTemplate=name
        )
    else:
        result["result"] = False
        result["comment"] = [
            f"gcp.compute.node_template#get(): {name} either resource_id or project, region and name should be specified."
        ]
        return result

    if not ret["result"]:
        result["comment"] += ret["comment"]
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result



"""State module for managing InstanceGroups."""

__contracts__ = ["resource"]

from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.state_operation_utils import StateOperations
from idem_gcp.tool.gcp.utils import zonal_absent

# prevent commit hook from removing the import
absent = zonal_absent

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str = None,
    zone: str = None,
    project: str = None,
    description: str = None,
    fingerprint: str = None,
    subnetwork: str = None,
    network: str = None,
    named_ports: List[
        make_dataclass(
            "NamedPort",
            [("name", str, field(default=None)), ("port", int, field(default=None))],
        )
    ] = None,
    request_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Create or update an instance group. Creates an instance group in the specified project using the parameters that are included in the request.


    Args:
        name(str):
            An Idem name of the resource.

        description(str, Optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.

        zone(str):
            The name of the zone where the instance group is located.

        fingerprint(str, Optional):
            The fingerprint of the named ports. The system uses this fingerprint to detect conflicts when multiple users change the named ports concurrently. Defaults to None.

        network(str, Optional):
            The URL of the network to which all instances in the instance group belong. If your instance has multiple network interfaces, then the network and subnetwork fields only refer to the network and subnet used by your primary interface (nic0). Defaults to None.

        subnetwork(str, Optional):
            The URL of the subnetwork to which all instances in the instance group belong. If your instance has multiple network interfaces, then the network and subnetwork fields only refer to the network and subnet used by your primary interface (nic0). Defaults to None.

        named_ports(List[Dict[str, Any]], Optional):
            Assigns a name to a port number. For example: {name: "http", port: 80} This allows the system to reference ports by the assigned name instead of a port number. Named ports can also contain multiple ports. For example: [{name: "app1", port: 8080}, {name: "app1", port: 8081}, {name: "app2", port: 8082}] Named ports apply to all instances in this instance group. . Defaults to None.
                * name (str, Optional): The name for this named port. The name must be 1-63 characters long, and comply with RFC1035.
                * port (int, Optional): The port number, which can be a value between 1 and 65535.

        project(str):
            Project ID for this request.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              gcp.compute.instance_groups.present:
                - name: value
                - zone: value
                - project: value
                - instance_group: value
    """

    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if ctx.get("wrapper_result"):
        result = ctx.get("wrapper_result")

    resource_body = {
        "description": description,
        "zone": zone,
        "name": name,
        "named_ports": named_ports,
        "network": network,
        "subnetwork": subnetwork,
    }

    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}
    operation = None
    if result["old_state"]:
        resource_id = result["old_state"].get("resource_id", None)
        resource_body["resource_id"] = resource_id

        # The fingerprint is required upon an update operation but in the time of creation the
        # resource still do not have fingerprint so, we cannot make it a required param for present method.
        resource_body["fingerprint"] = fingerprint or result["old_state"].get(
            "fingerprint"
        )

        # A dictionary of additional operations to perform on the object identified by the key
        # the values are tuples with the first element - the method to call, the second element - arguments,
        # third - the property is required in the end result
        patch_operations_dict = {
            "named_ports": (
                hub.tool.gcp.compute.instance_group.update_named_ports,
                (ctx, result["old_state"], named_ports),
                True,
            ),
        }

        state_operations = StateOperations(
            hub, "compute.instance_group", patch_operations_dict, result, resource_body
        )

        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"],
            {
                **resource_body,
            },
            "compute.instance_group",
            additional_exclude_paths=list(patch_operations_dict.keys()),
        )

        if changes:
            changed_non_updatable_properties = (
                hub.tool.gcp.resource_prop_utils.get_changed_non_updatable_properties(
                    "compute.instance_group", changes
                )
            )
            if changed_non_updatable_properties:
                result["result"] = False
                result["comment"].append(
                    hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                        "gcp.compute.instance_group",
                        name,
                        changed_non_updatable_properties,
                    )
                )
                result["new_state"] = result["old_state"]
                return result

        if not changes and not any(state_operations.changed_properties_dict.values()):
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(
                    "gcp.compute.instance_group", name
                )
            )
            result["new_state"] = result["old_state"]
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    "gcp.compute.instance_group", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            return result

        state_operations_ret = await state_operations.run_operations()

        if not state_operations_ret["result"] or not state_operations_ret.get(
            "new_state"
        ):
            result["result"] = False
            result["comment"] += state_operations_ret["comment"]
            return result

        result["new_state"] = state_operations_ret["new_state"]
        result["comment"].append(
            hub.tool.gcp.comment_utils.update_comment(
                "gcp.compute.instance_group", name
            )
        )
        return result

    else:
        # Create for test
        if ctx.get("test", False):
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.instance_group", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                resource_body
            )
            result["new_state"][
                "resource_id"
            ] = hub.tool.gcp.resource_prop_utils.construct_resource_id(
                "compute.instance_group", {**locals(), "instanceGroup": name}
            )
            return result

        # Create
        create_ret = await hub.exec.gcp_api.client.compute.instance_group.insert(
            ctx,
            name=name,
            project=project,
            zone=zone,
            request_id=request_id,
            body=resource_body,
        )
        if not create_ret["result"] or not create_ret["ret"]:
            result["result"] = False
            if create_ret["comment"] and next(
                (
                    comment
                    for comment in create_ret["comment"]
                    if "alreadyExists" in comment
                ),
                None,
            ):
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        "gcp.compute.instance_group", name
                    )
                )
            else:
                result["comment"] += create_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(create_ret["ret"]):
            operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.zone_operation"
        )
        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }

    return result


async def absent(
    hub,
    ctx,
    name: str = None,
    project: str = None,
    zone: str = None,
    request_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified instance group. The instances in the group are not deleted. Note that instance group must not belong to a backend service. Read Deleting an instance group for more information.

    Args:
        name(str):
            An Idem name of the resource.

        project(str):
            Project ID for this request.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

        zone(str):
            The name of the zone where the instance group is located.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              gcp.compute.instance_groups.absent:
                - name: value
                - project: value
                - instance_group: value
                - zone: value
    """
    # the method is handled via the recursive_contracts->call_absent
    raise NotImplementedError


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function. Retrieves the list of zonal instance group resources contained within the specified zone. For managed instance groups, use the instanceGroupManagers or regionInstanceGroupManagers methods instead.

    Args:
        page_token(str, Optional):
            Specifies a page token to use. Set `pageToken` to the `nextPageToken` returned by a previous list request to get the next page of results. Defaults to None.

        filter_(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = "Intel Skylake") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = "Intel Skylake") OR (cpuPlatform = "Intel Broadwell") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq "double quoted literal"` `(fieldname1 eq literal) (fieldname2 ne "literal")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`. Defaults to None.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy="creationTimestamp desc"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported. Defaults to None.

        max_results(int, Optional):
            The maximum number of results per page that should be returned. If the number of available results is larger than `maxResults`, Compute Engine returns a `nextPageToken` that can be used to get the next page of results in subsequent list requests. Acceptable values are `0` to `500`, inclusive. (Default: `500`). Defaults to 500.

        project(str):
            Project ID for this request.

        return_partial_success(bool, Optional):
            Opt-in for partial success behavior which provides partial results in case of failure. The default value is false. Defaults to None.

        include_all_scopes(bool, Optional):
            Indicates whether every visible scope for each scope type (zone, region, global) should be included in the response. For new resource types added after this field, the flag has no effect as new resource types will always include every visible scope for each scope type in response. For resource types which predate this field, if this flag is omitted or false, only scopes of the scope types where the resource type is expected to be found will be included. Defaults to None.

        zone(str):
            The name of the zone where the instance group is located.

        instance_group(str):
            The name of the instance group.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe gcp.compute.instance_groups
    """

    result = {}

    describe_ret = await hub.exec.gcp.compute.instance_group.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.instance_group {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")

        result[resource_id] = {
            "gcp.compute.instance_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



from pop.contract import ContractedContext


def pre(hub, ctx: ContractedContext):
    """
    If the endpoint_url wasn't specified in the func_ctx, then supply a default
    """
    func_ctx = ctx.get_argument("ctx")
    if "endpoint_url" not in func_ctx.acct:
        func_ctx.acct.endpoint_url = hub.exec.gcp_api.client.DEFAULT_ENDPOINT_URL


async def _ret(status: bool, ret=None, comment: str = ""):
    """
    Force the return to be a properly formatted coroutine
    """
    return {
        "comment": comment,
        "ret": ret,
        "status": status,
    }


async def call(hub, ctx: ContractedContext):
    """
    Catch all exceptions and return a False status if there was an error
    """
    try:
        ret = await ctx.func(*ctx.args, **ctx.kwargs)
        return _ret(ret=ret, status=True)
    except Exception as e:
        return _ret(comment=f"{e.__class__}: {e}", status=False)



"""State module for managing ServiceAccountKeys."""
from typing import Any
from typing import Dict

__contracts__ = ["resource"]
RESOURCE_TYPE_FULL = "gcp.iam.projects.service_accounts.key"


async def present(hub, ctx, name: str):
    """Present operation is not supported for this resource.

    Args:
        name(str):
            Idem name.

    Returns:
        .. code-block:: json

            {
                "result": False,
                "comment": "...",
            }

    """
    return {
        "result": False,
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": [
            hub.tool.gcp.comment_utils.no_resource_create_update_comment(
                RESOURCE_TYPE_FULL
            )
        ],
    }


async def absent(hub, ctx, name: str):
    """Absent operation is not supported for this resource.

    Args:
        name(str):
            Idem name.

    Returns:
        .. code-block:: json

            {
                "result": False,
                "comment": "...",
                "old_state": None,
                "new_state": None,
            }

    """
    return {
        "result": False,
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": [
            hub.tool.gcp.comment_utils.no_resource_create_update_comment(
                RESOURCE_TYPE_FULL
            )
        ],
    }


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieve the list of available keys for a given service account.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.iam.projects.service_accounts.key service_account_id=...
    """
    result = {}

    list_accounts_ret = await hub.exec.gcp.iam.projects.service_account.list(
        ctx, project=ctx.acct.project_id
    )

    if not list_accounts_ret["result"]:
        hub.log.debug(
            f"Could not describe {RESOURCE_TYPE_FULL} {list_accounts_ret['comment']}"
        )
        return result

    for resource in list_accounts_ret["ret"]:
        sa_resource_id = resource.get("resource_id")

        key_ret = await hub.exec.gcp.iam.projects.service_accounts.key.list(
            ctx, project=ctx.acct.project_id, sa_resource_id=sa_resource_id
        )

        if not key_ret["result"]:
            hub.log.debug(
                f"Could not describe {RESOURCE_TYPE_FULL} in {sa_resource_id}: {key_ret['comment']}"
            )
        else:
            for service_account_key in key_ret["ret"]:
                resource_id = service_account_key.get("resource_id")
                result[resource_id] = {
                    "gcp.iam.projects.service_accounts.key.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in service_account_key.items()
                    ]
                }

    return result



"""State module for managing Machine Images."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import global_absent

# prevent commit hook from removing the import
absent = global_absent


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    project: str = None,
    resource_id: str = None,
    source_instance: str = None,
    storage_locations: List[str] = None,
    source_disk_encryption_keys: List[
        make_dataclass(
            "SourceDiskEncryptionKey",
            [
                ("source_disk", str, field(default=None)),
                (
                    "disk_encryption_key",
                    make_dataclass(
                        "CustomerEncryptionKey",
                        [
                            ("kms_key_service_account", str, field(default=None)),
                            ("rsa_encrypted_key", str, field(default=None)),
                            ("kms_key_name", str, field(default=None)),
                            ("raw_key", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
            ],
        )
    ] = None,
    description: str = None,
    saved_disks: List[
        make_dataclass(
            "SavedDisk",
            [
                ("kind", str, field(default="compute#savedDisk")),
                ("storage_bytes_status", str, field(default=None)),
                ("architecture", str, field(default=None)),
                ("storage_bytes", str, field(default=None)),
                ("source_disk", str, field(default=None)),
            ],
        )
    ] = None,
    guest_flush: bool = None,
    machine_image_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kms_key_service_account", str, field(default=None)),
            ("rsa_encrypted_key", str, field(default=None)),
            ("raw_key", str, field(default=None)),
            ("kms_key_name", str, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates a machine image in the specified project using the data that is included in the request. If you are creating a new machine image to update an existing instance, your new machine image should use the same network or, if applicable, the same subnetwork as the original instance.

    Args:

        name(str):
            An Idem name of the resource.

        project(str, Optional):
            Project ID for this request.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        source_instance(str, Optional):
            The source instance used to create the machine image. You can provide this as a partial or full URL to the resource. For example, the following are valid values: - https://www.googleapis.com/compute/v1/projects/project/zones/zone /instances/instance - projects/project/zones/zone/instances/instance . Defaults to None.

        storage_locations(list[str], Optional):
            The regional or multi-regional Cloud Storage bucket location where the machine image is stored. Defaults to None.

        source_disk_encryption_keys(list[Dict[str, Any]], Optional):
            [Input Only] The customer-supplied encryption key of the disks attached to the source instance. Required if the source disk is protected by a customer-supplied encryption key. Defaults to None.
            * source_disk(str, Optional): URL of the disk attached to the source instance. This can be a full or valid partial URL. For example, the following are valid values: - https://www.googleapis.com/compute/v1/projects/project/zones/zone /disks/disk - projects/project/zones/zone/disks/disk - zones/zone/disks/disk
            * disk_encryption_key(Dict[str, Any], Optional): The customer-supplied encryption key of the source disk. Required if the source disk is protected by a customer-supplied encryption key.
                * kms_key_service_account(str, Optional): The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/
                * rsa_encrypted_key(str, Optional): Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem
                * kms_key_name(str, Optional): The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key
                * raw_key(str, Optional): Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

        description(str, Optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.

        saved_disks(list[Dict[str, Any]], Optional):
            An array of Machine Image specific properties for disks attached to the source instance. Defaults to None.
            * properties (Dict[str, Dict[str, Any]], optional): An instance-attached disk resource.
                * kind (str, optional): [Output Only] Type of the resource. Always compute#savedDisk for attached disks.
                * storageBytesStatus (str, optional): [Output Only] An indicator whether storageBytes is in a stable state or it is being adjusted as a result of shared storage reallocation. This status can either be UPDATING, meaning the size of the snapshot is being updated, or UP_TO_DATE, meaning the size of the snapshot is up-to-date.
                * architecture (str, optional): [Output Only] The architecture of the attached disk.
                * storageBytes (str, optional): [Output Only] Size of the individual disk snapshot used by this machine image.
                * sourceDisk (str, optional): Specifies a URL of the disk attached to the source instance.

        guest_flush(bool, Optional):
            [Input Only] Whether to attempt an application consistent machine image by informing the OS to prepare for the snapshot process. Defaults to None.

        machine_image_encryption_key(Dict[str, Any], Optional):
            Encrypts the machine image using a customer-supplied encryption key. After you encrypt a machine image using a customer-supplied key, you must provide the same key if you use the machine image later. For example, you must provide the encryption key when you create an instance from the encrypted machine image in a future request. Customer-supplied encryption keys do not protect access to metadata of the machine image. If you do not provide an encryption key when creating the machine image, then the machine image will be encrypted using an automatically generated key and you do not need to provide a key to use the machine image later. Defaults to None.
            * kms_key_service_account(str, Optional): The service account being used for the encryption request for the given KMS key. If absent, the Compute Engine default service account is used. For example: "kmsKeyServiceAccount": "name@project_id.iam.gserviceaccount.com/
            * rsa_encrypted_key(str, Optional): Specifies an RFC 4648 base64 encoded, RSA-wrapped 2048-bit customer-supplied encryption key to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rsaEncryptedKey": "ieCx/NcW06PcT7Ep1X6LUTc/hLvUDYyzSZPPVCVPTVEohpeHASqC8uw5TzyO9U+Fka9JFH z0mBibXUInrC/jEk014kCK/NPjYgEMOyssZ4ZINPKxlUh2zn1bV+MCaTICrdmuSBTWlUUiFoD D6PYznLwh8ZNdaheCeZ8ewEXgFQ8V+sDroLaN3Xs3MDTXQEMMoNUXMCZEIpg9Vtp9x2oe==" The key must meet the following requirements before you can provide it to Compute Engine: 1. The key is wrapped using a RSA public key certificate provided by Google. 2. After being wrapped, the key must be encoded in RFC 4648 base64 encoding. Gets the RSA public key certificate provided by Google at: https://cloud-certs.storage.googleapis.com/google-cloud-csek-ingress.pem
            * kms_key_name(str, Optional): The name of the encryption key that is stored in Google Cloud KMS. For example: "kmsKeyName": "projects/kms_project_id/locations/region/keyRings/ key_region/cryptoKeys/key
            * raw_key(str, Optional): Specifies a 256-bit customer-supplied encryption key, encoded in RFC 4648 base64 to either encrypt or decrypt this resource. You can provide either the rawKey or the rsaEncryptedKey. For example: "rawKey": "SGVsbG8gZnJvbSBHb29nbGUgQ2xvdWQgUGxhdGZvcm0="

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_resource_name:
              gcp.compute.machine_image.present:
                - project: project-name
                - source_instance: https://www.googleapis.com/compute/v1/projects/project-name/zones/zone-name/instances/instance-name
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = hub.tool.gcp.utils.get_project_from_account(ctx, project)

    if ctx.get("wrapper_result"):
        result = ctx.get("wrapper_result")

    request_body = {
        "storage_locations": storage_locations,
        "guest_flush": guest_flush,
        "source_instance": source_instance,
        "description": description,
        "name": name,
        "saved_disks": saved_disks,
        "machine_image_encryption_key": machine_image_encryption_key,
        "source_disk_encryption_keys": source_disk_encryption_keys,
    }

    operation = None
    plan_state = {"resource_id": resource_id, **request_body}
    plan_state = {k: v for (k, v) in plan_state.items() if v is not None}

    if result["old_state"]:
        resource_id = result["old_state"].get("resource_id", None)

        changes = hub.tool.gcp.utils.compare_states(
            result["old_state"], plan_state, "compute.machine_image"
        )

        if not changes:
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.up_to_date_comment(
                    "gcp.compute.machine_image", name
                )
            )
            result["new_state"] = result["old_state"]
            return result

        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_update_comment(
                    "gcp.compute.machine_image", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                plan_state
            )
            return result
        else:
            result["comment"].append(
                hub.tool.gcp.comment_utils.no_resource_update_comment(
                    "gcp.compute.machine_image"
                )
            )
            return result
    else:
        # Create machine image
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.machine_image", name
                )
            )
            result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(
                plan_state
            )
            result["new_state"][
                "resource_id"
            ] = hub.tool.gcp.resource_prop_utils.construct_resource_id(
                "compute.machine_image", {**locals(), "machineImage": name}
            )
            return result

        create_ret = await hub.exec.gcp_api.client.compute.machine_image.insert(
            ctx, project=project, body=request_body
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] += create_ret["comment"]
            return result

        if hub.tool.gcp.operation_utils.is_operation(create_ret["ret"]):
            operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.global_operation"
        )

        result["rerun_data"] = {
            "operation_id": operation_id,
            "old_state": result["old_state"],
        }
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of machine images that are contained within the specified project.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.machine_image
    """
    result = {}

    describe_ret = await hub.exec.gcp_api.client.compute.machine_image.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.machine_image {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"].get("items", []):
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.machine_image.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_peerings(
    hub,
    ctx,
    resource_id: str,
    current_peerings: List[Dict[str, Any]],
    new_peerings: List[Dict[str, Any]],
) -> Dict[str, Any]:
    result = {"result": False, "comment": []}

    current_peerings = current_peerings if current_peerings is not None else []
    new_peerings = new_peerings if new_peerings is not None else []

    existing_peering_names = {
        peering.get("name"): peering for peering in current_peerings
    }
    new_peering_names = {peering.get("name"): peering for peering in new_peerings}

    peerings_to_add = []
    peerings_to_update = []

    for new_peering in new_peerings:
        if new_peering.get("name") not in existing_peering_names:
            peerings_to_add.append(new_peering)
        else:
            peering_changes = hub.tool.gcp.utils.compare_states(
                {"peering": new_peering},
                {
                    "peering": next(
                        current_peering
                        for current_peering in current_peerings
                        if current_peering.get("name") == new_peering.get("name")
                    )
                },
                resource_type=None,
                additional_exclude_paths=["auto_create_routes"],
            )

            if peering_changes:
                peerings_to_update.append(new_peering)

    peerings_to_remove = [
        peering
        for peering in current_peerings
        if peering.get("name") not in new_peering_names
    ]

    # Remove peerings
    for peering in peerings_to_remove:
        remove_peering_request_body = {"name": peering.get("name")}

        remove_ret = await hub.exec.gcp_api.client.compute.network.removePeering(
            ctx, resource_id=resource_id, body=remove_peering_request_body
        )

        r = await hub.tool.gcp.operation_utils.await_operation_completion(
            ctx, remove_ret, "compute.network", "compute.global_operation"
        )
        if not r["result"]:
            result["comment"] += r["comment"]
            return result

    # Update peerings
    for peering in peerings_to_update:
        # We want to remove this property, as it will soon be deprecated.
        # exchange_subnet_routes should be used instead.
        peering.pop("auto_create_routes", None)

        # Ignoring this property, until figuring out why get does not return it and how to update it.
        peering.pop("peer_mtu", None)

        update_peering_request_body = {"network_peering": peering}

        update_ret = await hub.exec.gcp_api.client.compute.network.updatePeering(
            ctx, resource_id=resource_id, body=update_peering_request_body
        )

        r = await hub.tool.gcp.operation_utils.await_operation_completion(
            ctx, update_ret, "compute.network", "compute.global_operation"
        )
        if not r["result"]:
            result["comment"] += r["comment"]
            return result

    # Add peerings
    r = await hub.tool.gcp.compute.network.add_peerings(
        ctx, resource_id, peerings_to_add
    )

    if not r["result"]:
        result["comment"] += r["comment"]
        return result

    result["result"] = True
    return result


async def add_peerings(
    hub, ctx, resource_id: str, peerings_to_add: Dict[str, Any]
) -> Dict[str, Any]:
    result = {"result": False, "comment": []}

    for peering in peerings_to_add:
        # We want to remove this property, as it will soon be deprecated.
        # exchange_subnet_routes should be used instead.
        peering.pop("auto_create_routes", None)

        add_peering_request_body = {"network_peering": peering}

        add_ret = await hub.exec.gcp_api.client.compute.network.addPeering(
            ctx, resource_id=resource_id, body=add_peering_request_body
        )

        r = await hub.tool.gcp.operation_utils.await_operation_completion(
            ctx, add_ret, "compute.network", "compute.global_operation"
        )
        if not r["result"]:
            result["comment"] += r["comment"]
            return result

    result["result"] = True
    return result


def build_new_peerings_on_top_of_old(
    hub, old_peerings: List[Dict[str, Any]], new_peerings: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    if new_peerings:
        new_peerings = copy.deepcopy(new_peerings)
        # Ignoring this property, until figuring out why get does not return it and how to update it.
        for peering in new_peerings:
            peering.pop("peer_mtu", None)
    if not new_peerings or not old_peerings:
        return new_peerings

    new_peerings = new_peerings or []
    old_peerings = old_peerings or []

    new_peerings_body = []
    for new_peering in new_peerings:
        old_peering = next(
            peering
            for peering in old_peerings
            if new_peering.get("name") == peering.get("name")
        )
        if old_peering:
            new_peering_body = hub.tool.gcp.utils.create_dict_body_on_top_of_old(
                old_peering, new_peering
            )
        else:
            new_peering_body = new_peering

        new_peerings_body.append(new_peering_body)

    return new_peerings_body



"""State module for managing Cloud Key Management Service key rings."""
from typing import Any
from typing import Dict

__contracts__ = ["resource"]
RESOURCE_TYPE = "cloudkms.key_ring"
RESOURCE_TYPE_FULL = "cloudkms.projects.locations.key_rings"
GCP_RESOURCE_TYPE_FULL = "gcp.cloudkms.key_ring"


async def present(
    hub,
    ctx,
    name: str,
    key_ring_id: str = None,
    project_id: str = None,
    location_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Create a new KeyRing in a given Project and Location.

    Args:
        name(str:
            Idem name.

        key_ring_id(str, Optional):
            Key ring id.

        project_id(str, Optional:
            Project id.

        location_id(str, Optional):
            Location id.

        resource_id:
            Idem resource id. Formatted as

            `projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}`

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls
            key_ring_present:
              gcp.cloudkms.key_ring.present:
                - key_ring_id: idem-gcp
                  project_id: tango-gcp
                  location_id: us-east1

    """
    result = {
        "result": True,
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if hub.tool.gcp.resource_prop_utils.properties_mismatch_resource_id(
        RESOURCE_TYPE_FULL,
        resource_id,
        {
            "project_id": project_id,
            "location_id": location_id,
            "key_ring_id": key_ring_id,
        },
    ):
        result["comment"].append(
            hub.tool.gcp.comment_utils.properties_mismatch_resource_id_comment(
                RESOURCE_TYPE_FULL, name
            )
        )

    get_resource_only_with_resource_id = hub.OPT.idem.get(
        "get_resource_only_with_resource_id", False
    )
    if resource_id:
        old_get_ret = await hub.exec.gcp.cloudkms.key_ring.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"] or (
            not old_get_ret["ret"] and get_resource_only_with_resource_id
        ):
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        result["old_state"] = old_get_ret["ret"]
    elif not get_resource_only_with_resource_id:
        resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
            RESOURCE_TYPE_FULL,
            {
                "project_id": project_id,
                "location_id": location_id,
                "key_ring_id": key_ring_id,
            },
        )
        old_get_ret = await hub.exec.gcp.cloudkms.key_ring.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"]:
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        if old_get_ret["ret"]:
            result["old_state"] = old_get_ret["ret"]

    if result["old_state"]:
        resource_id = result["old_state"].get("resource_id", None)

        els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
            RESOURCE_TYPE_FULL, resource_id
        )
        if (
            els.get("project_id") != project_id
            or els.get("location_id") != location_id
            or els.get("key_ring_id") != key_ring_id
        ):
            result["result"] = False
            result["comment"].append(
                hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                    "gcp.cloudkms.key_ring",
                    key_ring_id,
                    ["project_id", "location_id", "key_ring_id"],
                )
            )
            return result
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_exists_comment(
                "gcp.cloudkms.key_ring", key_ring_id
            )
        )
        result["old_state"] = old_get_ret["ret"]
        result["new_state"] = old_get_ret["ret"]
        result["new_state"]["name"] = name
        return result

    if ctx["test"]:
        plan_state = {"resource_id": resource_id, "name": name}
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_create_comment(
                "gcp.cloudkms.key_ring", key_ring_id
            )
        )
        result["new_state"] = hub.tool.gcp.sanitizers.sanitize_resource_urls(plan_state)
        return result

    create_ret = (
        await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.create(
            ctx,
            parent=hub.tool.gcp.resource_prop_utils.construct_resource_id(
                "cloudkms.projects.locations",
                {"project_id": project_id, "location_id": location_id},
            ),
            key_ring_id=key_ring_id,
        )
    )
    if not create_ret["result"]:
        result["result"] = False
        result["comment"] += create_ret["comment"]
        return result
    result["comment"].append(
        hub.tool.gcp.comment_utils.create_comment("gcp.cloudkms.key_ring", key_ring_id)
    )
    result["old_state"] = {}
    resource_id = create_ret["ret"].get("resource_id")
    result["new_state"] = {"resource_id": resource_id, "name": name}
    return result


async def absent(
    hub,
    ctx,
    name: str,
) -> Dict[str, Any]:
    """Absent opreation is not supported for this resource.

    Args:
        name(str):
            Idem name.

    Returns:
        .. code-block:: json

            {
                "result": False,
                "comment": "...",
                "old_state": None,
                "new_state": None,
            }

    """
    return {
        "result": False,
        "name": name,
        "old_state": None,
        "new_state": None,
        "comment": [
            hub.tool.gcp.comment_utils.no_resource_delete_comment(
                "gcp.cloudkms.key_ring"
            )
        ],
    }


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieve the list of available key rings.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.cloudkms.key_ring
    """
    result = {}

    locations = await hub.exec.gcp.cloudkms.location.list(
        ctx, project=ctx.acct.project_id
    )
    if not locations["result"]:
        hub.log.debug(
            f"Could not describe gcp.cloudkms.key_ring in {ctx.acct.project_id} {locations['comment']}"
        )
        return {}

    for location in locations["ret"]:
        key_rings = await hub.exec.gcp.cloudkms.key_ring.list(
            ctx, location=location["resource_id"]
        )
        if not key_rings["result"]:
            hub.log.debug(
                f"Could not describe gcp.cloudkms.key_ring in {location['location_id']} {key_rings['comment']}"
            )
        else:
            for key_ring in key_rings["ret"]:
                resource_id = key_ring.get("resource_id")
                result[resource_id] = {
                    "gcp.cloudkms.key_ring.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in key_ring.items()
                    ]
                }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.generate.exec_param import ExecParam


async def should_operate(hub, execution_context: ExecutionContext) -> bool:
    return execution_context.get_exec_param_value(ExecParam.SANITIZE_PARAM)


async def operate(hub, execution_context: ExecutionContext) -> None:
    sanitized_params = {
        k: v for k, v in execution_context.method_params.items() if v is not None
    }
    execution_context.method_params = sanitized_params



"""Exec module for managing ServiceAccountKeys."""
from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext

__func_alias__ = {"list_": "list"}


async def list_(hub, ctx, sa_resource_id: str, project: str = None):
    r"""Lists every ServiceAccountKey that belongs to a specific project.

    Args:
        sa_resource_id(str):
            Resource id of the service account following the pattern projects/{project}/serviceAccounts/{id}.
        project(str, Optional):
            The resource name of the project associated with the service accounts.
    """
    execution_context = ExecutionContext(
        resource_type="iam.projects.service_accounts.key",
        method_name="list",
        method_params={"ctx": ctx, "name": sa_resource_id},
    )

    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def get(
    hub,
    ctx,
    project: str = None,
    service_account_id: str = None,
    key_id: str = None,
    resource_id: str = None,
):
    r"""Returns the specified ServiceAccountKey resource.

    Args:
        project(str, Optional):
            Project ID for this request.
        service_account_id(str, Optional):
            Email or unique_id of the service account in GCP.
        key_id(str, Optional):
            Id of the service account key in GCP.
        resource_id(str, Optional):
            An identifier of the service account key in idem. Defaults to None.
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    if service_account_id and key_id:
        project = hub.tool.gcp.utils.get_project_from_account(ctx, project)
        resource_id = (
            f"projects/{project}/serviceAccounts/{service_account_id}/keys/{key_id}"
        )
    elif not resource_id:
        result["result"] = False
        result["comment"] = [
            f"gcp.iam.projects.service_accounts.key#get(): either service_account_id and key_id or resource_id"
            f" should be specified."
        ]
        return result

    execution_context = ExecutionContext(
        resource_type="iam.projects.service_accounts.key",
        method_name="get",
        method_params={"ctx": ctx, "name": resource_id},
    )

    ret = await hub.tool.gcp.generate.generic_exec.execute(execution_context)

    result["comment"] += ret["comment"]
    if not ret["result"]:
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result
